// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *		HyperText Transport Protocol (HTTP) support
 *
 * This code defines a set of classes to carry out an HTTP transaction
 * as specified in a HTTP 1.1 document [RFC 2068]. Specifically, the code
 * provides classes that describe a single transaction, closures that submit
 * requests and closures to consume the returned content.
 * 
 * The closure classes ContentConsumer, RequestContent are abstract
 * classes (interfaces). They have to be specialized by the end user
 * depending on a particular request and a particular expected reply.
 * The header also defines a few service classes:
 * 	TypedFile - describes a file with an asscociated MIME type
 * 	
 * $Id: HTTP.h,v 1.5 1998/09/14 21:06:01 oleg Exp oleg $
 *
 ************************************************************************
 */

#if !defined(__GNUC__)
#pragma once
#else
#pragma interface
#endif

#ifndef __HTTP_h
#define __HTTP_h

#include "TCPstream.h"
#include "myenv.h"

#if defined(DELETE)
#undef DELETE
#endif

class RFC822Headers;

static const char CRLF [] = "\r\n";


			// Parameters that are necessary to open an HTTP pipe

struct HTTPConnection
{
public:
  char server_name[60];			// or an IP address
  int server_port;
  char proxy_name[60];
  int proxy_port;			// By default, the same as server's
  char resource_path[120];		// Abs Path to the resource

  char auth_credential[120];		// An authorization credential to use (or "")

  char identity_str[60];		// A string that identifies this User-Agent:

  bool q_use_proxy(void) const		{ return proxy_name[0] != '\0'; }
  const char * q_host_to_connect(void) const 
  			{ return q_use_proxy() ? proxy_name : server_name; }
  int q_port_to_connect(void) const 
  			{ return q_use_proxy() ? proxy_port : server_port; }

  bool q_send_auth(void) const		{ return auth_credential[0] != '\0'; }

public:
  class CTOR			// A Lazy constructor
  {
    friend class HTTPConnection;
  protected:
    virtual void fill_in(HTTPConnection& connection) const = 0;
  };
  HTTPConnection(const CTOR& lazy_constructor)
  	: server_port(80), proxy_port(80)
  	{ server_name[0] = '\0'; proxy_name[0] = '\0';
  	  resource_path[0] = '/'; resource_path[1] = '\0';
    	  auth_credential[0] = '\0';
	  xstrncpy(identity_str,"Unknown",sizeof(identity_str)-1);
    	  lazy_constructor.fill_in(*this);
  	}
};


		// This closure is used in a transaction to send a particular
		// kind of request. Specifically, a send_request_data() method
		// is called when a connection to an HHTP server
		// has been established and generic headers (Host, Authorization,
		// etc) have been written. The method is to send then additional
		// headers describing the request content (content-type:, etc)
		// and the content itself, if needed.

class RequestContent
{
protected:
  virtual void send_request_data (ostream& link)
  	{ link << CRLF;	// No data: an emty line - the end of HTTP headers
  	}
public:
  enum METHOD { GET=1, POST, PUT, HEAD, OPTIONS, DELETE };
  const METHOD method;
  
  RequestContent(METHOD _method) : method(_method) {}
  const char * q_method_str(void) const;
  friend ostream& operator << (ostream& link, RequestContent& file)
 	{ file.send_request_data(link); return link; }
};

struct ContentConsumer;


		// This is a class to prepare and carry out a single
		// HTTP transaction
class HTTPTransaction
{
public:
  enum TransStatus { OK, ReTry, Error };

private:
  HTTPTransaction(const HTTPTransaction&);	// private and unimplemented to
  void operator = (const HTTPTransaction&);	// prohibit copy-constructing
  
  HTTPConnection connection_parms;

		// HTTP status line, section 6.1 of the HTTP 1.1 document
		// Returns the status code, or EOF in case of an error
  static int read_response_status_line(istream& http_stream);
				// Receive the HTTP server's response
				// and process it
  TransStatus process_response(istream& link,ContentConsumer& content_consumer);

public:
  HTTPTransaction(const HTTPConnection::CTOR& connection_ctor)
  	: connection_parms(connection_ctor) {}
  TransStatus perform_transaction(RequestContent& request_content,
  				  ContentConsumer& consumer);
};

		// This closure is called upon to handle response data.
		// It is passed a stream containing the reply from the
		// HTTP server; the reply headers have beed read
		// (and passed separately)
		// The main operator() is called upon when an OK response
		// was received from the server. A handle_no_change()
		// method is called to handle a 304 response.
		// A handle_other_response() is called for any other
		// response code from the server. The latter function must
		// return a proper TransStatus.
struct ContentConsumer
{
  virtual bool operator () (istream& link, const RFC822Headers& headers) = 0;
  virtual bool handle_no_change(const RFC822Headers& headers);
  virtual HTTPTransaction::TransStatus
  	handle_other_response(const int resp_code,
			      istream& link, const RFC822Headers& headers);
};




			// A MIME-typed file
class TypedFile
{
  char file_name[255];
  char type[80];
  TypedFile(const TypedFile&);		// private and unimplemented to
  void operator = (const TypedFile&);	// prohibit copy-constructing
public:
  TypedFile(const char file_name [], const char file_type []);
  friend ostream& operator << (ostream& link, const TypedFile& file);
};

#endif
