// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 * 			Dealing with MIME Entities
 *
 * This code is to read an entity from a message stream, that is, to
 * convert entities from an externalized to an internalized form, so
 * to speak.
 * To this end, the code reads entity's header fields, locates and makes
 * sense of the relevant (starting with 'content-') headers, loads the
 * body performing necessary conversions. Although at present no conversions
 * are performed.
 *
 * The biggest hurdle is handling of a composite entities, which may consist
 * of separate parts (each of them can be a multipart entity in turn).
 * The grammar of the mult-part messages is tricky, as RFC 2046 warns
 * the implementors. Here's this grammar from RFC 2046, quoted here
 * for easy reference.
 *
 * dash-boundary := "--" boundary
 *                    ; boundary taken from the value of boundary parameter
 *                    ; of the Content-Type field.
 *
 * multipart-body :=	[preamble CRLF]
 *			dash-boundary transport-padding CRLF
 *			body-part *encapsulation
 *			close-delimiter transport-padding
 *			[CRLF epilogue]
 *
 * transport-padding := *LWSP-char
 *
 * encapsulation := 	delimiter transport-padding
 * 			CRLF body-part
 *
 * delimiter := 	CRLF dash-boundary
 *
 * close-delimiter := 	delimiter "--"
 *
 * preamble := discard-text
 *
 * epilogue := discard-text
 *
 * discard-text := 	*(*text CRLF) *text
 * 			; May be ignored or discarded.
 *
 * body-part := 	MIME-part-headers [CRLF *OCTET]
 *			; Lines in a body-part must not start
 *			; with the specified dash-boundary and
 *			; the delimiter must not appear anywhere
 *			; in the body part.
 *
 * The RFC 2046 especially notes that
 * "The CRLF preceding the boundary delimiter line is conceptually
 * attached to the boundary so that it is possible to have a part that
 * does not end with a CRLF (line  break).  Body parts that must be
 * considered to end with line breaks, therefore, must have two CRLFs
 * preceding the boundary delimiter line, the first of which is part of
 * the preceding body part, and the second of which is part of the
 * encapsulation boundary."
 *
 * $Id: MIMEEntity.cc,v 4.3 1998/04/20 18:43:29 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifdef __GNUC__
#pragma implementation
#endif

#include <fstream.h>
#include "MIMEEntity.h"
#include <stdio.h>
#include <ctype.h>

			// HTTP character types
			// Section "2.2 Basic Rules" of the HTTP 1.1
			// document (see the reference above)
char HTTPCharType::char_types [128] =
{
// nul soh stx  etx  eot   enq    ack    bel
   0,  0,   0,   0,    0,    0,   0,      0,	// 0x00 through 0x07
// bs  ht   nl   vt    np    cr   so     si
   0,  0,   0,   0,    0,    0,   0,      0,	// 0x08 through 0x0f
// dle dc1  dc2  dc3   dc4  nak   syn   etb
   0,  0,   0,   0,    0,    0,   0,      0,	// 0x10 through 0x17
// can em  sub  esc   fs    gs   rs     us
   0,  0,   0,   0,    0,    0,   0,      0,	// 0x18 through 0x1f
// sp   !    "    #     $      %      &      '
   0, _token,0,_token,_token,_token,_token,_token, // 0x20 through 0x27
// (    )    *     +    ,    -      .    /
   0,   0,_token,_token,0,_token,_token, 0,	// 0x28 through 0x2f
// 0      1       2     3      4      5      6      7
 _token,_token,_token,_token,_token,_token,_token,_token,// 0x30 through 0x37
// 8      9       :     ;      <      =      >      ?
 _token,_token,   0,    0,     0,     0,     0,     0,	// 0x38 through 0x3f
// @      A      B     C      D      E      F      G
   0,  _token,_token,_token,_token,_token,_token,_token,// 0x40 through 0x47
// H      I       J     K      L      M      N      O
 _token,_token,_token,_token,_token,_token,_token,_token,// 0x48 through 0x4f
// P      Q       R     S      T      U      V      W
 _token,_token,_token,_token,_token,_token,_token,_token,// 0x50 through 0x57
// X      Y       Z     [      \      ]      ^      _
 _token,_token,_token,  0,     0,     0,   _token,_token,// 0x58 through 0x5f
// `      a      b     c      d      e      f      g
 _token,_token,_token,_token,_token,_token,_token,_token,// 0x60 through 0x67
// h      i       j     k      l      m      n      o
 _token,_token,_token,_token,_token,_token,_token,_token,// 0x68 through 0x6f
// p      q       r     s      t      u      v      w
 _token,_token,_token,_token,_token,_token,_token,_token,// 0x70 through 0x77
// x      y       z     {      |      }      ~     del
 _token,_token,_token,  0,  _token,   0,   _token,  0	 // 0x78 through 0x7f
};

//------------------------------------------------------------------------
//			Parsing an HTTP header
//
// According to the HTTP 1.1 document,
//
// "HTTP header fields, which include general-header (section 4.5), request-
// header (section 5.3), response-header (section 6.2), and entity-header
// (section 7.1) fields, follow the same generic format as that given in
// Section 3.1 of RFC 822 [9]. Each header field consists of a name
// followed by a colon (":") and the field value. Field names are case-
// insensitive. The field value may be preceded by any amount of LWS,
// though a single SP is preferred. Header fields can be extended over
// multiple lines by preceding each extra line with at least one SP or HT.
// Applications SHOULD follow "common form" when generating HTTP
// constructs, since there might exist some implementations that fail to
// accept anything beyond the common forms.
//
//      message-header = field-name ":" [ field-value ] CRLF
//
//       field-name     = token
//       field-value    = *( field-content | LWS )
//
//       field-content  = <the OCTETs making up the field-value
//                        and consisting of either *TEXT or combinations
//                        of token, tspecials, and quoted-string>
// <EOQ>
// This version of parser relaxes the rules above when it comes to line
// terminating. The grammar for message headers requires CRLF at the
// end of every header field. We take this rule a bit leniently and accept
// any conventional line terminator: a single LF, a single CR, or a CRLF
// combination.

		// This is the function that reads and parses a header
		// field, which is supposed to start from the current
		// position in in_stream. Continuation lines, if
		// present, will be read too.
		// On success, the program queues a RFC822Headers::Header object
		// describing the header, and returns a "Continue" code.
		// If an empty line is detected (which signifies the
		// end of headers), an Empty_Line code is returned.
		// On error, the function returns IO_Error or Syntax_Error
		// All activity is logged into stream 'clog'

		// The parsing algorithm is implemented as a Finite State
		// machine
RFC822Headers::Digesting_Result RFC822Headers::read_a_header(istream& in_stream)
{
  Logger clog;

  StringCollection::Mark start_name = buffer.new_string();
  StringCollection::Mark start_value = StringCollection::Mark::invalid_mark;
  
  enum Status			// current state of the FSM
  	{Start_of_Name, In_Name,
  	 Start_Value, In_Value, CRLF_in_val, In_Leading_CRLF }
  status = Start_of_Name;
  
  for(;;)
  {
    char c;
    if( !in_stream.get(c).good() )
      return (clog << "Input error or unexpected EOF while reading a header"),
      		  IO_Error;
    clog << c;

    switch(status)
    {
      case Start_of_Name:		// Initial state: about to read the name
   	   if( HTTPCharType::is_token_char(c) )
   	     buffer << c, status = In_Name;
   	   else if( c == '\r' )
   	     status = In_Leading_CRLF;
   	   else if( c == '\n' )
   	     in_stream.putback(c), status = In_Leading_CRLF;
   	   else
   	     return Syntax_Error;
   	   continue;
   	   
      case In_Name:			// Accumulating characters of the name
      	   if( HTTPCharType::is_token_char(c) )
      	     buffer << c;		// Accumulate token chars
      	   else if( c == ':' )
      	     buffer << ends, start_value = buffer.new_string(),
      	     status = Start_Value;
      	   else
      	     return Syntax_Error;
      	   continue;
      	   
      case Start_Value:			// Skipping LWS before the value
           if( HTTPCharType::is_lws_char(c) )
             continue;
           status = In_Value;		// Falling through
           
      case In_Value:			// accumulating characters of 'Value'
           if( c == '\r' )
             status = CRLF_in_val;
   	   else if( c == '\n' )
   	     in_stream.putback(c), status = CRLF_in_val;
           else
             buffer << c;
           continue;
           
      case CRLF_in_val:			// Got CR while reading the value
           if( c != '\n' )		// The previous character could've been
             in_stream.putback(c);	// a single CR terminating a line
           if( HTTPCharType::is_lws_char(in_stream.peek()) )
           {
             status = In_Value;		// The next line starts with LWS:
             continue;			// it is a continuation line
           }
           else
           {
             buffer << ends;
             break;
           }
      
      case In_Leading_CRLF:		// Got CR right upfront, while reading name
           if( c != '\n' )		// The previous character could've been
             in_stream.putback(c);	// a single CR terminating a line
           return Empty_Line;
   
      default:
  	    assert(/*Unknown status*/0);
    }
    break;
  }
//  assert( buffer.good() );
  assert( start_value );
  add_assoc(start_name,start_value);
  return Continue;
}

			// Constructor: reads all the header lines from
			// the in_stream, until an empty line or
			// error
RFC822Headers::RFC822Headers(istream& in_stream)
{
  for(;;)
  {
    Digesting_Result result = read_a_header(in_stream);
    if( result == Continue )
     continue;
    if( result != Empty_Line )
      Logger() << "Error " << (int)result << " while reading headers";
    break;
  }
  buffer << ends;
}

			// Dump the contents of the headers
ostream& operator << (ostream& os, const RFC822Headers& headers)
{
  return os << "\n\n---> RFC822 headers\n"
	    << static_cast<const Dictionary&>(headers) << endl;
}


			// Build an enitity stream from an input character
			// stream, which is supposed to contain the entity's
			// body only. The headers must be supplied separately
MIMEEntityStream::MIMEEntityStream(const RFC822Headers& _headers,
				   istream& body_stream,
			           const char _boundary[])
  : en_stream(body_stream), headers(_headers), boundary(_boundary)
{
}

  			// Check to see if the current symbol is
  			// CR, LF, or CR followed by LF
  			// If so, scan through it and return true
bool MIMEEntityStream::seen_CRLF(void)
{
  switch( en_stream.peek() )
  {
    case '\n':
 	 return en_stream.get(), true;
 	 
    case '\r':
  	  en_stream.get();			// CR is seen: eat it
 	  if( en_stream.peek() == '\n' )	// If LF follows eat it too
 	    en_stream.get();
 	  return true;
 	  
    default:
 	 return false;
  }
  return false;		// unnecessary stmt, just to make gcc happy 
}


  			// Check to see if there are two dashes at
  			// the current position in the stream.
  			// If so, scan through them and return true.
  			// Otherwise, the current position in the
  			// stream remains unchanged
bool MIMEEntityStream::seen_dash_dash(void)
{
  if( en_stream.peek() != '-' )
    return false;
  register const int c = en_stream.get();
  if( en_stream.peek() == '-' )
    return en_stream.get(), true;	// Scan the second dash, if seen
  else if( c != EOF )
    en_stream.putback(c);
  return false;
}

  			// Skip linear-white spaces (space or HT)
  			// The stream will be positioned at the first
  			// not-LWS character, or at EOF
void MIMEEntityStream::skip_lwspaces(void)
{
  while( HTTPCharType::is_lws_char(en_stream.peek()) )
    en_stream.get();
}

  			// Skip the stream all the way through
  			// 	CRLF -- boundary
  			// Return true if succeeded, in that case
  			// the stream is positioned right after the
  			// last character of the boundary
  			// Return false if EOF or other error occurred
bool MIMEEntityStream::skip_through_dash_boundary(const char boundary [])
{
  while( en_stream.good() )
  {
    if( ! seen_CRLF() )
    {
      en_stream.get();
      continue;
    }
    if( !( en_stream.get() == '-' && en_stream.get() == '-' ) )
      continue;
      
    for(register const char * p = boundary; *p != '\0'; p++)
      if( en_stream.get() != *p )
        continue;
    return true;
  }
  return false;
}


			// Load the entity from the stream and have it
			// handled (with a given handler)
			// Return the indication how successful the handling was
			// The stream is positioned after the whole entity
			// was read; if the boundary was specified, it is scanned
			// through, too.
bool MIMEEntityStream::operator >> (const EntityHandler& handler)
{
  iststream ct_stream(headers,"Content-Type");
  
  if( !ct_stream )
    return (Logger() << "No Content-Type: header given"), false;

				// Parse the content-type header's value
  ContentType content_type(ct_stream);
  Logger() << "Parsed Content-type string" << content_type;

  if( !content_type.q_multipart() )
  {					// Read a discrete part
    MIMEDiscreteEntity discrete_entity(headers,content_type,en_stream,boundary);
    return discrete_entity.is_valid() &&
           ( const_cast<EntityHandler&>(handler)(discrete_entity) || 
             (discrete_entity.leave_content_file(), false) );
  }
  else
  
  			// We transform the multipart entity grammar
  			// (see above) into the following:
  			// multipart-body := [preamble CRLF] dash-boundary
  			// *( transport-padding CRLF body-part CRLF dash-boundary )
  			// "--" transport-padding [CRLF epilogue]
  {
    char nested_boundary[120];
    iststream bd_stream(content_type,"boundary");
    if( !bd_stream )
      return (Logger() << "No boundary= parameter in the multipart content"),
      false;
    bd_stream.getline_upto(nested_boundary,sizeof(nested_boundary),"");
    en_stream.putback('\n');
    if( !skip_through_dash_boundary(nested_boundary) )
      return false;		// Skip through the preamble and 1st boundary
     
      // the stream is positioned right after a dash-boundary terminating
      // the previous part or the preamble. EOF is ok at this point
      // (as an implicit close boundary)
    while( !seen_dash_dash() && !eof() )
    {
      if( skip_lwspaces(), !seen_CRLF() )	// skip transport padding
        return (Logger() << "Garbage after the boundary"),
        	false;

        	// Now we are positioned to read the encapsulated body part
       RFC822Headers nested_headers(en_stream);
       if( !(MIMEEntityStream(nested_headers,en_stream,
       			      nested_boundary) >> handler) )
         return (Logger() << "Reading of a composite entity terminated "
         	"due to a problem handling the part read before"), false;
    }
    if( eof() )
      return true;
    if( boundary[0] != '\0' )
      skip_through_dash_boundary(boundary);
    else
      skip_lwspaces();			// skip the epilogue
    return true;
  }
  return false;		// unnecessary stmt, just to make gcc happy 
}

MIMEDiscreteEntity::MIMEDiscreteEntity(const RFC822Headers& _headers,
  		     const ContentType& _content_type,
  		     istream& en_stream, const char boundary[])
	: headers(_headers), content_type (_content_type),
	  valid_flag(false)
{
  assert( tmpnam(body_file_name) );
  if( !en_stream.good() )
  {
    Logger() << "The entity stream is not in a good state: can't load entity";
    return;
  }
  
  ofstream content_file(body_file_name,ios::out|ios::binary);

  const int length = headers.lookup("Content-Length",-1);
  if( length == -1 )
    if( boundary[0] == '\0' )
  	valid_flag = receive_content_data(content_file,en_stream); // No length, no boundary
    else
  	valid_flag = receive_content_data(content_file,en_stream,boundary);
    else if( boundary[0] == '\0' )
  	valid_flag = receive_content_data(content_file,en_stream,length);
    else
  	valid_flag = receive_content_data(content_file,en_stream,length,boundary);
  	
  if( valid_flag && !content_file.good() )
    Logger() << "error writing the entity body file",
    valid_flag = false;
}

MIMEDiscreteEntity::~MIMEDiscreteEntity(void)
{
  if( body_file_name && body_file_name[0] != '\0' )
    remove(body_file_name);
}


			// Do not remove the content file
  			// upon destruction
void MIMEDiscreteEntity::leave_content_file(void)
{
  Logger() <<  "Error handling the content, leaving the content file "
    << body_file_name;
  body_file_name[0] = '\0';
}


  			// Receive the content data into a content_file
  			// until the EOF
bool MIMEDiscreteEntity::receive_content_data(ostream& content_file,
					      istream& en_stream)
{
  Logger clog;
  clog << "Content-Length: wasn't found; can't report on progress..." << endl;
  
  content_file << en_stream.rdbuf();
  
  if( en_stream.fail() )
     return (clog << "error reading entity body from the stream"), false;
    
  return true;
}

  			// Receive the content data into a content_file
  			// with a given length
bool MIMEDiscreteEntity::receive_content_data(ostream& content_file,
		istream& en_stream, const int length)
{
  Logger clog;

  for(register int i=0; i<length; i++)
  {
    register const int c = en_stream.get();
    if( c == EOF )
      return (clog << "premature EOF or other error reading the content"), false;
     content_file.put((char)c);
  }
  return true;
}

  			// Receive the content data into a content_file
  			// until the given delimiter has been seen, or EOF
  			// has occurred
  			// The delimiter is
  			// 	CRLF -- boundary
bool MIMEDiscreteEntity::receive_content_data(ostream& content_file,
			istream& en_stream, const char boundary[])
{
  Logger clog;
  clog << "Content-Length: wasn't found; can't report on progress..." << endl;
  
  char lookahead_buf[76];
  char * lap = lookahead_buf;
  assert( strlen(boundary) < sizeof(lookahead_buf) - 5 );
  
  for(;;)
  {
    if( lap != lookahead_buf )		// Empty a look-ahead buffer
    {
      for(register const char * p = lookahead_buf; p < lap; p++)
        content_file.put(*p);
      lap = lookahead_buf;
    }
    register const int c = en_stream.get();
    if( c == EOF )
      break;
    switch(c)
    {
      case '\n':
	   *lap++ = '\n';
    	   break;
    	   
      case '\r':
	   *lap++ = '\r';
	   if( en_stream.peek() == '\n' )
	     *lap++ = en_stream.get();
    	   break;

      default:
	   content_file.put((char)c);
	   continue;
    }
    				// seen CRLF
    if( en_stream.peek() == '-' )
      *lap++ = en_stream.get();
    else
      continue;
    if( en_stream.peek() == '-' )
      *lap++ = en_stream.get();
    else
      continue;
    				// seen CRLF --
    register const char * p = boundary ;
    while( *p != '\0' )
      if( en_stream.peek() == *p++ )
        *lap++ = en_stream.get();
      else
        break;
    if( *p == '\0' )		// Read the whole boundary
      break;
  }
  
  if( en_stream.fail() )
    return (clog << "error reading entity body from the stream"), false;
    
  return true;
}

  			// Receive the content data into a content_file
  			// with a given length. After that, check for and scan
  			// the delimiter
  			// The delimiter is
  			// 	CRLF -- boundary
  			// Note, it's OK to get EOF after the body was
  			// completely read (in that case, the delimiter is
  			// implied)
bool MIMEDiscreteEntity::receive_content_data(ostream& content_file,
		istream& en_stream, const int length, const char boundary[])
{
  if( !receive_content_data(content_file,en_stream,length) )
    return false;
  switch( en_stream.get() )
  {
    case '\n':
 	 break;
 	 
    case '\r':
 	 if( en_stream.peek() == '\n' )
 	   en_stream.get();
 	 break;
 	 
    case EOF:
 	 return true;
 	 
    default:
 	 Logger() << "CRLF was not found after having read the body";
 	 return false;
  }
  if( !( en_stream.get() == '-' && en_stream.get() == '-' ) )
    return (Logger() << "-- was not found after having read the body"), false;
    
  for(register const char * p = boundary; *p != '\0'; p++)
    if( en_stream.get() != *p )
      return (Logger() << "boundary was not found after having read the body"),
      false;
  return true;
}


//------------------------------------------------------------------------
//			Parsing of parameterized tokens
// a string like
//	ext-token *( ";" parameter )
//
// where ext-token is a token possibly containing "/" symbols, and
// parameters may follow the ext-token in the form of attribute/value
// pairs.
//
//       parameter      = attribute "=" value
//       attribute      = token
//       value          = token | quoted-string
// The ext-token and parameter attribute names are case-insensitive.
// Parameter values may or may not be case-sensitive, depending on the
// semantics of the parameter name. Linear white space (LWS) MUST NOT be
// used in the ext-token, nor between an attribute and its value
// For example, a Content-Type HTTP and MIME header follows the structure
// above
//

ParameterizedToken::ParameterizedToken(const char first_key_name [],
				       iststream& source)
{
  StringCollection::Mark start_key = buffer.new_string();
  buffer << first_key_name << ends;
  StringCollection::Mark start_value = buffer.new_string();
  enum Status			// current state of the FSM
  	{start_of_ExtToken, in_ExtToken,
  	 looking_for_Attrib, start_Attrib, in_Attrib,
  	 looking_for_Value,  start_Value, in_Quoted_Value, in_Token_Value,
  	 Exit, Err_bad_char }
  status = start_of_ExtToken;

  for(;;)
  {
    const char c = source.get();

    switch(status)
    {
      case start_of_ExtToken:		// Initial state: about to read the exttok
   	   if( c == ' ' || c == '\t' )	// skip leading whitespaces
   	     continue;
   	   if( c == EOF || c == ';' || c == '/' )
   	     buffer << "[Empty Token]";
   	   status = in_ExtToken;	// Falling through to accumulate the token
   	   
      case in_ExtToken:			// Accumulating characters of ctype
      	   if( HTTPCharType::is_token_char(c)  || c == '/' )
      	     { buffer << c; continue; }		// Accumulate token chars
      	   buffer << ends;
      	   assert( start_key && start_value );
    	   add_assoc(start_key,start_value);
    	   start_key = start_value = StringCollection::Mark::invalid_mark;
    	   
    	   				// Falling through to handle a non-token
    	   				// character
    	   status = looking_for_Attrib;
      	   
      case looking_for_Attrib:		// Skipping LWS before the ";" if any
     	   if( c == ' ' || c == '\t' )
     	     continue;
     	   if( c == ';' )
     	     { status = start_Attrib; continue; }
     	   status = c == EOF ? Exit : Err_bad_char;
     	   break;
           
      case start_Attrib:		// About to read the attr
   	   if( c == ' ' || c == '\t' )	// skip leading whitespaces
   	     continue;
           start_key = buffer.new_string();
   	   if( c == EOF || c == '=' )
   	     buffer << "[Empty Attrib]";
   	   status = in_Attrib;		// Falling through to accumulate the token
   	   
      case in_Attrib:			// Accumulating characters of the attrib
      	   if( HTTPCharType::is_token_char(c) )
      	     { buffer << c; continue; }		// Accumulate token chars
      	   buffer << ends;
    	   
    	   				// Falling through to handle a non-token
    	   				// character
    	   status = looking_for_Value;
      	   
      case looking_for_Value:		// Skipping LWS before the '='
      	   if( c == ' ' || c == '\t' )
      	     continue;
	   else if( c == '=' )
     	     { status = start_Value; continue; }
     	   status = Err_bad_char;
     	   break;
           
      case start_Value:			// Skipping LWS before the value
   	   if( c == ' ' || c == '\t' )	// skip leading whitespaces
   	     continue;
           start_value = buffer.new_string();
           if( c == '"' )
             { status = in_Quoted_Value; continue; }
           status = in_Token_Value;	// Falling through
           
      case in_Token_Value:		// accumulating characters of 'Value'
      	   if( HTTPCharType::is_token_char(c) )
      	     { buffer << c; continue; }		// Accumulate token chars
      	   buffer << ends;
      	   assert( start_key && start_value );
    	   add_assoc(start_key,start_value);
    	   start_key = start_value = StringCollection::Mark::invalid_mark;
    	   
    	   if( c == EOF )
    	    { status = Exit; break; }
    	   status = looking_for_Attrib;
    	   source.putback(c);
    	   continue;
           
      case in_Quoted_Value:		// Accumulate a quoted string
           if( c != '"' && c != EOF )
      	     { buffer << c; continue; }		// Accumulate string chars
           if( c == EOF )
             buffer << "[Missing terminating quote]";
      	   buffer << ends;
      	   assert( start_key && start_value );
    	   add_assoc(start_key,start_value);
    	   start_key = start_value = StringCollection::Mark::invalid_mark;
    	   
    	   status = looking_for_Attrib;
    	   continue;
   
      default:
  	    assert(/*Unknown status*/0);
    }
    
    if( status == Err_bad_char )
    {
      Logger clog;
      clog << "Bad character '";
      if( c == EOF ) clog << "EOF"; else clog << c;
      clog << "']";
      break;
    }
    else if( status == Exit )
      break;
    else
      assert(/*Unknown status*/0);
  }
}

const char ContentType::MediaType_Key [] = "__content-type-key";

bool ContentType::q_multipart(void) const
{
  return iststream(*this,MediaType_Key).begins_with_ci("multipart/").good();
}
