/*
 *----------------------------------------------------------------------------
 *			HTTPFS / File System API glue
 *
 * This source code is an adapter between a HTTPFS client framework and
 * a POSIX file system API. This code implements functions that act
 * as regular POSIX open(), close(), stat(), and lstat() calls. The functions
 * in the present code accept the same parameters as the corresponding POSIX
 * API calls, and return the same result (including errno).
 * However, if a file with a "http://"-prefixed name is being accessed,
 * the interposed functions forward the corresponding operations to the
 * HTTPFS client framework (see mchttp-client.h, mchttp-client.cc).
 * 
 * $Id: httpfs-client.cc,v 1.2 1999/05/02 22:49:12 oleg Exp oleg $
 *----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <string.h>
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <utime.h>
#include <assert.h>



#include "mchttp-client.h"

#include "myenv.h"
#include "Logger.h"
#include <fstream.h>

		// Prefix of HTTP extended file names
static const char HTTP_EXTENDED_FNAME_PREFIX [] = "http://";

		// Given "http://..." filename, return the path name
		// without the "http:/" prefix
		// Otherwise, return 0
static inline const char * strip_http_prefix(const char filename [])
{
  return 
  strncmp(filename,HTTP_EXTENDED_FNAME_PREFIX,
	  strlen(HTTP_EXTENDED_FNAME_PREFIX)) == 0 ?
  filename + 6 : 0;
}
	// Declarations of exported (extended) POSIX file API functions
	// that are over-ridden and intercepted by this package
extern "C" {
int httpe_open(const char *filename, const int mode, const int mask);
int httpe_close(const int fildes);
int httpe_stat(const char *path, struct stat *buf);
int httpe_lstat(const char *path, struct stat *buf);
}


static char Curr_directory [512];


static VNode_list VNodes(16);


//---------------------------------------------------------------------------
// 			Opening and closing of files

		// Check to see if 'filename' to open has the
		// HTTP_EXTENDED_FNAME_PREFIX. If so, get the HTTPFS
		// client to open a VNODE for that file. Otherwise,
		// the regular open(2) will be used.

int httpe_open(const char *filename, const int mode, const int mask)
{
  Logger() << "Request: open: '" << filename << "', mode " << mode;
  const char * vpath = strip_http_prefix(filename);
  if( !vpath) 
    return open(filename,mode,mask);

  VHandle handle = VNodes.open(vpath,mode);
  if( !(bool)handle )
  {
    Logger() << "V-node opening failed";
    return -1;
  }
  return VNodes.get_cache_handle(handle);
}

		// First try to close 'fildes' as if it is
		// a fildes of a Vnode cache file. If this fails,
		// the regular close(2) will be used.
int httpe_close(const int fildes)
{
  if( VNodes.close_cache_handle(fildes) )
    return 0;
  else
    return close(fildes);
}


//---------------------------------------------------------------------------
// 			Getting File Status

class SetStatusInfo : public ContentConsumer
{
  struct stat &target_buffer;
public:
  bool operator () (istream& link, const RFC822Headers& headers)
  { FileStatus status_info(headers);
    if( !(bool)status_info )
      return (clog << "invalid stat"), false;
     target_buffer = status_info;
     return true;
   }
   SetStatusInfo(struct stat & _buf) : target_buffer(_buf) {}
};

int httpe_lstat(const char *filename, struct stat *buf)
{
  Logger() << "Request: lstat: " << filename;
  const char * vpath = strip_http_prefix(filename);
  if( !vpath) 
    return lstat(filename,buf);
    
  HTTPTransaction http_trans(
			    VFSNameParser::vfs_name_to_httpfs(vpath));
  return
    http_trans.perform_transaction(
	RequestWithCommand(RequestContent::HEAD,"lstat"),SetStatusInfo(*buf))
       == HTTPTransaction::OK ?
       0 :
	(Logger() << "Transaction error", errno = EIO, -1);
}

int httpe_stat(const char *filename, struct stat *buf)
{
  Logger() << "Request: stat: " << filename;
  const char * vpath = strip_http_prefix(filename);
  if( !vpath) 
    return stat(filename,buf);
 
  FileStatus cached_status = VNodes.get_stat(vpath);
  if( (bool)cached_status )
  {
    Logger() << "sending cached status";
    *buf = cached_status;
    return 0;
  }
  HTTPTransaction http_trans(
			    VFSNameParser::vfs_name_to_httpfs(vpath));
  return
    http_trans.perform_transaction(
	RequestWithCommand(RequestContent::HEAD,"stat"),SetStatusInfo(*buf))
       == HTTPTransaction::OK ?
       0 :
	(Logger() << "Transaction error", errno = EIO, -1);
}

//---------------------------------------------------------------------------
// 			Scanning of a Directory

#if 0
void do_opendir (void)
{
  char *arg;

  rpc_get (msock, RPC_STRING, &arg, RPC_END);

    		// make sure that a dir name ends in a slash
  char * dir_name = arg;
  if( dir_name[strlen(dir_name)-1] != '/' )
  {
    const int old_len = strlen(arg);
    dir_name = (char *)realloc(dir_name,old_len+2);
    assert( dir_name );
    dir_name[old_len] = '/';
    dir_name[old_len+1] = '\0';
  }
  VHandle handle = VNodes.open(dir_name,O_RDONLY);
  if( (bool)handle )
    send_status(handle.get_raw_handle(),0);
  else
    Logger() << "V-node opening failed",
    send_status(-1,errno == 0 ? EIO : errno);
  free(dir_name);
}

/* Sends the complete directory listing, as well as the stat information */
void do_readdir (void)
{
    int    handle_raw;
 
    rpc_get (msock, RPC_INT, &handle_raw, RPC_END);
    
    Logger clog;
    clog << "readdir for " << handle_raw;
    VHandle handle(handle_raw);
    const char * const cache_file = VNodes.get_cache_file(handle);
    clog << "Cache file for " << handle << " is >" 
	 << cache_file << "<";
    if( cache_file == 0 || *cache_file == '\0' )
    {
      rpc_send (msock, RPC_INT, 0, RPC_END);
      return;
    }
    ifstream dir_stream(cache_file);
    char buffer[100];
    if( !dir_stream.good() )
    {
      clog << "Failed to open the cache file " << cache_file;
      rpc_send (msock, RPC_INT, 0, RPC_END);
      return;
    }
    while( !dir_stream.eof() )
    {
      dir_stream.get(buffer,sizeof(buffer)-1,'/');
	  //clog << "read '" << buffer << "'" << endl;
      bool good_flag = dir_stream.get() == '/' || buffer[0] != '\0';
      if( !good_flag )
	xstrncpy(buffer,"io-error",sizeof(buffer)-1);
      rpc_send (msock, RPC_INT, strlen(buffer), RPC_END);
      rpc_send (msock, RPC_BLOCK, strlen(buffer), buffer, RPC_END);
      if( !good_flag )
      {
	clog << "invalid name"; send_status(-1, EIO);
	return;
      }
      dir_stream.getline(buffer,sizeof(buffer)-1,'\n');
      if( !dir_stream.good() )
      {
	send_status(-1, EIO);
	return;
      }
      FileStatus status(buffer);
	  //clog << "read stat: " << buffer << endl;
      if( !(bool)status )
      {
	clog << "invalid stat"; send_status(-1, EIO);
	return;
      }
      send_status(0,0);
      send_stat_info(status);
      dir_stream.peek();		// Forces EOF
    }
    rpc_send (msock, RPC_INT, 0, RPC_END);
}

void do_closedir (void)
{
  int raw_handle;

  rpc_get (msock, RPC_INT, &raw_handle, RPC_END);
  if( !VNodes.close(VHandle(raw_handle)) )
    Logger() << "error closing vnode #" << raw_handle;
}
#endif

/* }}} */

/* {{{ Operations with one and two file name argument */

#if 0
void do_chdir (void)
{
    char *file;
    int  status;

    rpc_get (msock, RPC_STRING, &file, RPC_END);
    
#if 0
    status = chdir (file);
#else
    Logger() << "Request: chdir: " << file;
    if( strlen(file) < sizeof(Curr_directory) )
    {
      strcpy(Curr_directory,file);
      status = errno = 0;
    }
    else
      status = -1, errno = ENAMETOOLONG;
#endif
    send_status (status, errno);
    free (file);
}

void do_rmdir (void)
{
    char *file;
    int  status;

    rpc_get (msock, RPC_STRING, &file, RPC_END);
    
#if 0
    status = rmdir (file);
#else
    Logger() << "Request: rmdir: " << file;
    status = -1, errno = EIO;
#endif
    send_status (status, errno);
    free (file);
}

void do_mkdir (void)
{
    char *file;
    int  mode, status;

    rpc_get (msock, RPC_STRING, &file, RPC_INT, &mode, RPC_END);
#if 0
    status = mkdir (file, mode);
#else
    Logger() << "Request: mkdir: '" << file << "' mode " << mode;
    status = -1, errno = EIO;
#endif
    
    send_status (status, errno);
    free (file);
}

void do_mknod (void)
{
    char *file;
    int  mode, dev, status;

    rpc_get (msock, RPC_STRING, &file, RPC_INT, &mode, RPC_INT, &dev, RPC_END);
    
#if 0
    status = mknod (file, mode, dev);
#else
    Logger() << "Request: mknod: '" << file << "' mode " << mode
             << " dev " << dev;
    status = -1, errno = EACCES;
#endif
    send_status (status, errno);
    free (file);
}

void do_readlink (void)
{
    char *file;

    rpc_get (msock, RPC_STRING, &file, RPC_END);
    Logger() << "Request: readlink: " << file;
    HTTPTransaction http_trans(
		VFSNameParser::vfs_name_to_httpfs(file));
    free (file);
    
    struct SendSLinkTarget : public ContentConsumer
    {
      bool operator () (istream& link, const RFC822Headers& headers)
     { 
       char buffer [2048];
       link.get(buffer,sizeof(buffer)-1,'\n');
       if( link.get() != '\n' || buffer[0] == '\0' )
	 return false;
       send_status(strlen(buffer),0);
       rpc_send (msock, RPC_STRING, buffer, RPC_END);
       return true;
     }
     SendSLinkTarget(void) {}
    };

    if( http_trans.perform_transaction(
	RequestWithCommand(RequestContent::GET,"readlink"),SendSLinkTarget())
       != HTTPTransaction::OK )
    {
      Logger() << "Transaction error";
      send_status(-1,EIO);
      return;
    }
#if 0
    n = readlink (file, buffer, 2048);
    send_status (n, errno);
    if (n >= 0) {
        buffer [n] = 0;
	rpc_send (msock, RPC_STRING, buffer, RPC_END);
    }
    free (file);
#endif
}

void do_unlink (void)
{
    char *file;
    int  status;
    
    rpc_get (msock, RPC_STRING, &file, RPC_END);
#if 0
    status = unlink (file);
#else
    Logger() << "Request: unlink: " << file;
    status = -1, errno = EIO;
#endif
    send_status (status, errno);
    free (file);
}

void do_rename (void)
{
    char *f1, *f2;
    int  status;
    
    rpc_get (msock, RPC_STRING, &f1, RPC_STRING, &f2, RPC_END);
#if 0
    status = rename (f1, f2);
#else
    Logger() << "Request: rename: '" << f1 << "' to '" << f2 << "'";
    status = -1, errno = EIO;
#endif
    send_status (status, errno);
    free (f1); free (f2);
}

void do_symlink (void)
{
    char *f1, *f2;
    int  status;
    
    rpc_get (msock, RPC_STRING, &f1, RPC_STRING, &f2, RPC_END);
#if 0
    status = symlink (f1, f2);
#else
    Logger() << "Request: symlink: '" << f1 << "' to '" << f2 << "'";
    status = -1, errno = EACCES;
#endif
    send_status (status, errno);
    free (f1); free (f2);
}

void do_link (void)
{
    char *f1, *f2;
    int  status;
    
    rpc_get (msock, RPC_STRING, &f1, RPC_STRING, &f2, RPC_END);
#if 0
    status = link (f1, f2);
#else
    Logger() << "Request: link: '" << f1 << "' to '" << f2 << "'";
    status = -1, errno = EACCES;
#endif
    send_status (link (f1, f2), errno);
    free (f1); free (f2);
}
#endif

/* }}} */

/* {{{ Misc commands */

#if 0
void do_chmod (void)
{
    char *file;
    int  mode, status;
    
    rpc_get (msock, RPC_STRING, &file, RPC_INT, &mode, RPC_END);
#if 0
    status = chmod (file, mode);
#else
    Logger() << "Request: chmod: '" << file << "' mode " << mode;
    Logger() << "Replying OK but not actually changing the mode...";
//    status = -1, errno = EPERM;
    status = 0, errno = 0;
#endif
    send_status (status, errno);
    free (file);
}

void do_chown (void)
{
    char *file;
    int  owner, group, status;
    
    rpc_get (msock, RPC_STRING, &file,RPC_INT, &owner, RPC_INT,&group,RPC_END);
#if 0
    status = chown (file, owner, group);
#else
    Logger() << "Request: chown: '" << file << "' owner " << owner
    	     << " group " << group;
    status = -1, errno = EPERM;
#endif
    send_status (status, errno);
    free (file);
}

#endif
