/*
 *----------------------------------------------------------------------------
 *			HTTPFS Client / MCFS - HTTPFS Adapter
 *
 * This code implements the HTTPFS client framework, which carries out
 * HTTP transactions with a HTTPFS server and maintains a local cache of
 * fetched files and directory listings: VNodes. A file being opened for
 * reading or modification is first fetched from a server in a GET
 * transaction. However, if the file is already in cache, a conditional
 * GET request is issued to verify that the cached copy is up-to-date,
 * and reload it if not. When a file is being opened for writing,
 * an additional Pragma: header is included in the GET request to inform
 * the server of the open mode: O_RDWR, O_WRONLY, O_CREAT, O_EXCL, O_TRUNC
 * or O_APPEND. The server may then create, truncate, or lock the resource.
 * A response from the server is translated into the result of the open()
 * call. Reads and writes to the opened file are then directed to the
 * local copy. On close(), if the local copy has been modified, it is
 * written back using PUT.
 *
 * Status inquiries, e.g., stat(), lstat(), readlink(), etc., are
 * implemented by submitting a HEAD request.  A Pragma: request header
 * tells the server which particular status information about the resource
 * is requested.
 *
 * Scanning of a directory -- opendir(), readdir(), closedir() -- is
 * similar to accessing a file: a GET request is issued for a directory
 * URI, and the resulting directory listing is locally cached.
 * 
 * See MCHFS-server.pl for a server part of HTTPFS. Source code
 * file httpfs-client.cc provides a POSIX-like interface for this
 * HTTPFS client framework; Source code file mcserv.cc offers another
 * access to this framework, as a glue between the HTTPFS and
 * a Midnight Commander's MCFS file system.
 * 
 * A mapping between the file system API and HTTP requests and responses:
   
    "open" file flags mode
	Try to locate a valid VNode for the corresponding file. Otherwise,
   	allocate a new one. Issue a HTTPFS request:
		GET file
		Pragma: httpfs="preopen-xxxx"
		If-modified-since: yyyyy
	where xxxx encodes the open() flags: O_RDONLY, O_RDWR, O_WRONLY,
		O_CREAT, O_EXCL and O_TRUNC
	The HTTPFS server thus gets a chance to check if the remote resource
	can indeed be retrieved, modified, created or truncated.
	See MCHFS-server.pl for more details.
	If the file is being opened for modification (not O_RDONLY), 
	a 'dirty' field of the VNodeFile is set to true.
	The "If-modified-since: yyyyy" header is present if the cached VNode
	was located; yyyyy is the value of a VNode::last_checked in a HTTP
	date format.
	The cache file is then open, and its real handle is returned
	as the result of an open() operation.
    "close" real-handle
    	Locate an open VNode whose opened cache file has a handle
	equal to the 'real-handle'.
    	if the file has been opened for writing (VNodeFile::dirty is set),
	upload the contents of the cache file by issuing a
		PUT file
	HTTPFS request.
    "read" real-handle count
    	Do a regular read(2) operation on the 'real-handle' (which is
	a handle of an open local cache file). No HTTP request issued
    "write" real-handle count
    	Do a regular write(2) operation on the 'real-handle' (which is
	a handle of an open local cache file). No HTTP request issued.
    "lseek" real-handle offset whence
   	Do a regular lseek(2) operation on the 'real-handle' (which is
	a handle of an open local cache file). No HTTP request issued.
	
    "stat" filename-URL
    	First we check to see if there is a valid VNode for the given
	filename-URL (possibly with a '/' appended, in case it turns out
	to be a directory). If such a VNode is found, its cached status
	information is returned, without bothering the HTTPFS server.
	Otherwise, issue a HTTPFS request:
    		HEAD file
		Pragma: httpfs="stat"
	
	The server returns the status of a given resource in a 
        ETag: response header
		ETag: '"' status-info '"'
	where status-info is a string of 11 numbers separated by a single
        space: dev ino mode nlink uid gid size atime mtime ctime blocks
	where 
		dev      device number of filesystem
                ino      inode number
                mode     file mode  (type and permissions)
                nlink    number of (hard) links to the file
                uid      numeric user ID of file's owner
                gid      numeric group ID of file's owner
                size     total size of file, in bytes
                atime    last access time since the epoch
                mtime    last modify time since the epoch
                ctime    inode change time (NOT creation time!) since the epoch
                blocks   actual number of blocks allocated
	all numbers but 'mode' are decimal numbers; mode is an octal number.
	see struct stat and 'stat' entry in man perlfunc

	Note the status-info is a "hard validator" - that is, it is a unique
	identification for a resource (file or directory). Indeed, should
	the file gets modified, at least its modification timestamp will
	change. That's why the status-info is being delivered in a ETag: a
	field which is designated by the HTTP standard to carry
	resource identifiers.
	
     "lstat " file
        HTTPFS request:
    		HEAD file
		Pragma: httpfs="lstat"
	The answer is the same as that of the "stat" above
	
    "fstat" real-handle
    	return the file status information from the corresponding real-handle
	(which is a handle of an open local cache file).
	
    "opendir" dir-name
    	dir-name always ends in a slash (or is made to end in a slash).
    	Look for a valid VNodeDir with this dir-name, or allocate a new one.
	Issue a request
		GET dir-name
		If-modified-since: yyyyy
	The "If-modified-since: yyyyy" header is present if the cached VNode
	was located; yyyyy is the value of a VNode::last_checked in a HTTP
	date format.
	The server returns the listing of the directory: for each
	directory entry
	(including . and ..) the server writes a line
		name/status-info
	see above for status-info
		Last-Modified: last-modif time of the directory
	This information is written in a cache file of the VNodeDir.
	The VHandle of this VNode is returned to a MCFS client.
    
    "readdir" dir-handle
    	The 'dir-handle' is supposed to be a VHandle of a VNodeDir.
	This VNode is located and its cache file is parsed and sent to
	a MCFS client (as a sequence of "name" stat-for-the-name
	for each name in the directory in question).
 	No HTTP request issued.

     
    "closedir" dir-handle
    	The 'dir-handle' is supposed to be a VHandle of a VNodeDir, which
	is thus closed.
	No HTTP request issued.
	
    "chdir" dir-name
    	simply store the dir-name in a global field Curr_directory
	No HTTP request issued.

     "rmdir" dir-name
        HTTPFS request:
     		DELETE dir-name
	where dir-name is guaranteed to end in a slash
	
     "mkdir" dir-name mode
        HTTPFS request:
     		PUT dir-name
	where dir-name is guaranteed to end in a slash
	
     "mknod" file mode dev
     	not implemented, always returns errno = EACCES
	
     "readlink" file-name
        HTTPFS request:
        	HEAD file-name
		Pragma: httpfs="readlink"

     "unlink" file-name
        HTTPFS request:
      		DELETE file-name
	
     "rename" old-name new-file-name
      	??? file-name

     "symlink" old-name new-file-name
      	not implemented, always returns errno = EACCES

     "link" old-name new-file-name
      	not implemented, always returns errno = EACCES

    "chmod" file-name mode
     	not implemented, but returns as if the request was successful
	(to make a CMFS client happy).

    "chown" file owner group
     	not implemented, always returns errno = EPERM
    
    "utime"
     	not implemented, always returns errno = EPERM
 *
 * $Id: mchttp-client.cc,v 2.5 1999/05/21 18:23:39 oleg Exp oleg $
 *----------------------------------------------------------------------------
 */


#if defined(__GNUC__)
#pragma implementation
#endif

#include "mchttp-client.h"

#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#include <errno.h>

static const char This_Client_Name [] = "HTTPFS/1.1"; // User-Agent header

#include "HTTP.h"
#include "MIMEEntity.h"
#include "myenv.h"
#include <fstream.h>
#include <time.h>
#include <assert.h>


/*
 *-------------------------------------------------------------------------
 *	Contacting an HTTP server (e.g., a MCHFS-server.pl CGI script)
 */

		// Parse the vfs_name and fill out the HTTPConnection
		// connection parameters
void VFSNameParser::fill_in(HTTPConnection& connection) const
{
  Logger() << "parsing: " << vfs_name;
  		// will be over-written later if we succeed in parsing
  xstrncpy(connection.server_name,"VFSNameParser-error",
	   sizeof(connection.server_name)-1);

  if( vfs_name[0] != '/' )
  {
    Logger() << "vfs name '" << vfs_name << "' must start with a slash";
    return;
  }
  const char * const res_name = strchr(vfs_name+1,'/');
  if( res_name == 0 )
  {
    Logger() << "vfs name '" << vfs_name << "': needs / after the hostname";
    return;
  }
  xstrncpy(connection.resource_path,res_name,sizeof(connection.resource_path)-1);
  const char * proxy_del = strchr(vfs_name,'^');
  if( proxy_del > res_name )
    proxy_del = 0;

  if( proxy_del )
  {
    const char * const proxy_name = proxy_del + 1;
    const char * p = strchr(proxy_name,':');
    if( p== 0 || p > res_name )
      p = res_name;
    if( p < res_name )
      connection.proxy_port = atoi(p+1);
    if( connection.proxy_port > 0 && p > proxy_name )
      {
        assert( p-proxy_name < sizeof(connection.proxy_name) - 1);
        xstrncpy(connection.proxy_name,proxy_name,p-proxy_name);
      }
    else
    {
      xstrncpy(connection.proxy_name,"invalid-proxy-name",
	   sizeof(connection.proxy_name)-1);
      Logger() << "vfs name '" << vfs_name << "': can't figure the proxy name";
      return;
    }
   }

   {
    const char * const hname_end = proxy_del ? proxy_del : res_name;
    const char * p = strchr(vfs_name,':');
    if( p== 0 || p > hname_end )
      p = hname_end;
    if( p < hname_end )
      connection.server_port = atoi(p+1);
    if( connection.server_port > 0 && p > vfs_name+1 )
    {
      assert( p-vfs_name-1 < sizeof(connection.server_name) - 1);
      xstrncpy(connection.server_name,vfs_name+1,p-vfs_name-1);
    }
    else
    {
      xstrncpy(connection.server_name,"invalid-server-name",
	   sizeof(connection.server_name)-1);
      Logger() << "vfs name '" << vfs_name << "': can't figure the server name";
      return;
    }
   }

  xstrncpy(connection.identity_str,This_Client_Name,
	   sizeof(connection.identity_str)-1);
}


		// Parse the status-info string of 11 numbers
		// see the title comments for more details.
		// If the parsing succeeds, we build a fully filled struct stat
		// If failed to parse the string, we invalidate
		// the object and return false
bool FileStatus::load_from_status_info(const char status_info [])
{
  st_rdev = 0; 
  		// unlike struct stat fields, we know that all the following
  		// are int's, and we can safely use %d scanf format
  int dev, ino, mode, nlink, uid, gid, size, atime, mtime, ctime, blocks;
  if( sscanf(status_info,"%d %d %o %d %d %d %d %d %d %d %d",
	&dev,&ino,&mode,&nlink,&uid,&gid,&size,
	&atime,&mtime,&ctime,&blocks) != 11 )
  {
    Logger() << "invalid status-info string: " << status_info;
    invalidate();
    return false;
  }
  st_dev = dev;
  st_ino = ino;
  st_mode = mode;
  st_nlink = nlink;
  st_uid = uid;
  st_gid = gid;
  st_size = size;
  st_atime = atime;
  st_mtime = mtime;
  st_ctime = ctime;
  st_blocks = blocks;
  return true;
}


		// Look up an ETag header and parse the stat info
		// If the parsing succeeds, we build a fully filled struct stat
		// If failed to parse the string, we invalidate
		// the object and return false
bool FileStatus::load_from_etag(const RFC822Headers& headers)
{
  const StringCollection::Mark etag_mark = headers.lookup("ETag");
  if( !etag_mark )
    return (Logger() << "No ETag: found"), invalidate(), false;
  iststream istr(headers,etag_mark);
  if( istr.get() != '"' )
    return invalidate(), false;
  char buffer[256];
  istr.getline_upto(buffer,sizeof(buffer)-1,"\"");
  if( istr.get() != '"' )
    return invalidate(), false;
  return load_from_status_info(buffer);
}

/*
 *-------------------------------------------------------------------------
 *			Vnodes and V-node cache
 * These data structures are intended to hold all the information we
 * have received from a HTTPFS server concerning some particular file or
 * directory.
 */

VNode::VNode(const char _name [])
  : last_checked(0), cache_handle(-1), ref_count(0)
{
  assert( strlen(_name) < sizeof(name) );
  strcpy(name,_name);
  assert( tmpnam(cache_file) );
}

VNode::~VNode(void)
{
  assert_consistency();
  assert( ref_count == 0 );
  remove(cache_file);
}

bool VNode::open(const int open_flags)
{
  assert( ref_count == 0 );	// At present, a VNode can be opened by only one
  ref_count++;			// client
  return true;
}

bool VNode::open_cache_handle(const int open_flags)
{
  assert( (bool)(*this) && cache_handle < 0 );
  cache_handle = ::open(cache_file,open_flags);
  return cache_handle < 0 ? (invalidate(), false) : true;
}

bool VNode::close(void)
{
  assert( ref_count > 0 );
  if( cache_handle >= 0 ) 
    ::close(cache_handle);
  cache_handle = -1;
  ref_count--;
  return true;
}

  			// Make sure the VNode is in a logically consistent
  			// state
void VNode::assert_consistency(void) const
{
  assert( name[0] != '\0' );
  assert( cache_file[0] != '\0' );
  assert( ref_count >= 0 && ref_count <= 1 );
  if( (bool)(*this) )
  {
    assert( (bool)stat_info );
    assert( last_checked > 0 );
    if( cache_handle >= 0 ) assert( ref_count == 1 );
  }
  else
  {
    assert( last_checked == 0 );
    assert( ref_count == 0 );
    assert( cache_handle == -1 );
    assert( !q_in_use() );
  }
}

			// Take the content HTTP server sends us
 			// and load it into the local cache file
bool VNode::CacheLoad::operator () (istream& link, const RFC822Headers& headers)
{ 
  vnode.stat_info = FileStatus(headers);
  if( !(bool)vnode.stat_info )
    return (Logger() << "invalid stat"), false;
  vnode.last_checked = time(0);
  ofstream cache_stream(vnode.cache_file);
  assert( cache_stream.good() );
  cache_stream << link.rdbuf();
  assert( cache_stream.good() );
  if( !(link.peek(),link.eof()) )
    return (Logger() << "apparently problem reading data from the link"),
    vnode.invalidate(),
    false;
  return true;
}

			// Handle specifically a 201 response code from
			// a HTTPFS server: an indication that we need to
			// create a cache file of zero size
HTTPTransaction::TransStatus
VNode::CacheLoad::handle_other_response(const int resp_code,
			      istream& link, const RFC822Headers& headers)
{
  if( resp_code != 201 )
    return ContentConsumer::handle_other_response(resp_code,link,headers);
  Logger() << "Received response code 201: truncating the cache file";
  const int fh = ::creat(vnode.cache_file,0666);
  if( fh < 0 || ::fstat(fh,&vnode.stat_info) < 0 )
    return (Logger() << "failed to creat() the cache file"),
    vnode.invalidate(),
    HTTPTransaction::Error;
  ::close(fh);
  vnode.last_checked = time(0);
  return HTTPTransaction::OK;
}

			// Handle specifically a "201 Created" response code from
			// a HTTPFS server and update the file_status
HTTPTransaction::TransStatus
VNode::PUTRefresh::handle_other_response(const int resp_code,
			      istream& link, const RFC822Headers& headers)
{
  if( resp_code != 201 )
    return ContentConsumer::handle_other_response(resp_code,link,headers);
  Logger() << "Received response code 201: updating the stat_info";
  vnode.stat_info = FileStatus(headers);
  if( !(bool)vnode.stat_info )
    return (Logger() << "invalid stat"), false;
  vnode.last_checked = time(0);
  return HTTPTransaction::OK;
}


		/* The name of a directory always ends in '/'
		 */

class VNodeDir : public VNode
{
public:
  VNodeDir(const char name []);
  bool open(const int open_flags);
};

VNodeDir::VNodeDir(const char _name []) : VNode(_name)
{
  Logger() << "Creating new VNodeDir: " << _name 
	   << " cached into " << get_cache_file();
  assert( _name[strlen(_name)-1] == '/' );
}

			// Opening a dir node; This fn ignores open_flags
			// as a directory may only be opened for reading
bool VNodeDir::open(const int open_flags)
{
  assert( !q_in_use() );
  Logger() << "Opening VNodeDir: " << get_name() << endl;
  HTTPTransaction http_trans(
	VFSNameParser::vfs_name_to_httpfs(get_name()));
  if( http_trans.perform_transaction(
       RequestContent_IfModif(RequestContent::GET,get_last_checked()),CacheLoad(*this))
       != HTTPTransaction::OK )
    return Logger() << "transaction failed", errno=EIO,
    invalidate(), false;
  else
   return VNode::open(open_flags);
}


class VNodeFile : public VNode
{
  bool dirty;
public:
  VNodeFile(const char name []);
  bool open(const int open_flags);
  bool close(void);
};

VNodeFile::VNodeFile(const char _name [])
	: VNode(_name), dirty(false)
{
  Logger() << "Creating new VNodeFile: " << _name 
	   << " cached into " << get_cache_file();
}


bool VNodeFile::open(const int open_flags)
{
  assert( !q_in_use() );
  Logger() << "Opening VNodeFile: " << get_name() << endl;
  
  class ReqOpen : public RequestContent_IfModif
  {
    char preopen_chars[5];
    void set_preopen_chars(const int open_flags)
    {
      preopen_chars[0] = 
	( (open_flags & O_ACCMODE) == O_RDONLY ||
	  (open_flags & O_ACCMODE) == O_RDWR ) ? 'R' : '-';
      preopen_chars[1] = 
      (open_flags & O_ACCMODE) == O_RDONLY ? '-' :
         (  (open_flags & O_APPEND) ? 'A' : 'W' );
      preopen_chars[2] = (open_flags & O_CREAT) ?
        ( (open_flags & O_EXCL) ? 'X' : 'C' ) : '-';
      preopen_chars[3] = (open_flags & O_TRUNC) ? 'T' : '-';
      preopen_chars[4] = '\0';
    }
   public:
    void send_request_data (ostream& link)
    { link << "Pragma: httpfs=\"preopen-"  << preopen_chars << '"' << CRLF;
      RequestContent_IfModif::send_request_data(link);
    }
    ReqOpen(const int open_flags, const time_t modif_time) :
     RequestContent_IfModif(RequestContent::GET,modif_time)
     {set_preopen_chars(open_flags);
      Logger() << "preopening with flags:" << preopen_chars; }
  };
  
  HTTPTransaction http_trans(
	VFSNameParser::vfs_name_to_httpfs(get_name()));
  if( http_trans.perform_transaction(
       ReqOpen(open_flags,get_last_checked()),CacheLoad(*this))
       != HTTPTransaction::OK )
    return Logger() << "transaction failed",
           errno = EIO,
           invalidate(), false;
    const int cache_open_flags = open_flags & ~O_CREAT & ~O_EXCL;
    dirty = (open_flags & O_ACCMODE) != O_RDONLY;
    if( open_cache_handle(cache_open_flags) && VNode::open(open_flags) )
      return assert_consistency(), true;
    else
      return Logger() << "Opening failed", errno = EIO, invalidate(), false;
}


		// A dummy consumer
struct DummyConsumer : public ContentConsumer
{
  bool operator () (istream& link, const RFC822Headers& headers)
  { cout << "\nReceieved content\n" << link.rdbuf() << endl;
    return true; }
};

bool VNodeFile::close(void)
{
  if( !VNode::close() )
    return false;
  if( !dirty )
    return true;
  struct put_requestor : public RequestContent
  {
    TypedFile file;
    void send_request_data (ostream& link) { link << file; }
    put_requestor(const char file_name [])
    	: RequestContent(RequestContent::PUT),
      	  file(file_name,"application/octet-stream")
      	{ }
  };
  
  HTTPTransaction http_trans(
	VFSNameParser::vfs_name_to_httpfs(get_name()));
  if( http_trans.perform_transaction(
       put_requestor(get_cache_file()),PUTRefresh(*this))
       != HTTPTransaction::OK )
    return Logger() << "transaction failed",
           errno = EIO,
           invalidate(), false;
  return true;
}


VNode_list::VNode_list(const int _size)
	: size(_size), vnode_array(new (VNode *) [_size])
{
  memset((void*)vnode_array,0,sizeof(vnode_array[0])*size);
}


			// Lookup a vnode with a given name. Return VListPtr
			// describing the result. If vnode was found,
			// VListPtr.handle will be its handle, and
			// *(VListPtr.vpp) would be that vnode's pointer
			// If a vnode was not found, lookup() returns
			// a VListPtr to the first free node slot.
			// In that case, VListPtr.vpp is zero.
			// If the name can't be located and the list is
			// full, VListPtr.handle will be invalid
VNode_list::VListPtr::VListPtr(const VNode_list& list, const char name [])
	: vpp(0)
{
  register VNode **cvpp = list.vnode_array;
  for(register int i=1; i<=list.size; i++,cvpp++)
  {
    if( *cvpp == 0 )
    {
      handle = i, vpp = cvpp;
      continue;
    }
    if( (*cvpp)->match(name) )
    {
      handle = i, vpp = cvpp;
      return;
    }
   }
}

			// Make a free vnode by finding an appropriate
			// victim and killing it.
			// An eligible victim is an invalid vnode or
			// vnode with a ref_count zero.
			// Return a VListPtr pointing to a slot with
			// the killed node (if successful). Note, this
			// slot pointer is guaranteed to be 0. 
			// If all vnode slots are full with valid
			// and referenced vnodes, return a VListPtr
			// with an invalid handle
VNode_list::VListPtr::VListPtr(VNode_list& list, PURGE_GET_FREE)
	: vpp(0)
{
  		// first search for invalid vnodes
  register VNode **cvpp = list.vnode_array;
  register int i;
  for(i=1, cvpp=list.vnode_array; i<=list.size; i++,cvpp++)
  {
    if( *cvpp == 0 )
      continue;
    if( !(bool)(**cvpp) )
    {
      delete *cvpp;
      *cvpp = 0;
      handle = i; vpp = cvpp;
      return;
    }
  }
  		// if all vnodes in the list are valid, search for
  		// a non-referenced vnode
  for(i=1, cvpp=list.vnode_array; i<=list.size; i++,cvpp++)
  {
    if( *cvpp != 0 && (*cvpp)->q_in_use()  )
      continue;
    if( *cvpp != 0 )
      delete *cvpp;
    *cvpp = 0;
    handle = i; vpp = cvpp;
    return;
  }
}

			// Lookup a vnode with a given handle. Return VListPtr
			// describing the result. If the handle was invalid
			// or points to un unexisting node, OR that vnode
			// is invalid or not in use, return
			// an invalid VListPtr and set errno to EBADF
VNode_list::VListPtr::VListPtr(const VNode_list& list, const VHandle _handle)
	: handle(_handle), vpp(0)
{
  if( !(bool)handle || handle.get_raw_handle() > list.size )
  {
    Logger() << "Vnode " << handle << " out of range";
    errno = EBADF;
    return;
  }
  register VNode **cvpp = &(list.vnode_array[handle.get_raw_handle()-1]);
  if( *cvpp == 0 )
  {
    Logger() << "Vnode " << handle << " does not exist";
    errno = EBADF;
    return;
  }
  if( !(bool)(**cvpp) || !(**cvpp).q_in_use() )
  {
    Logger() << "Vnode " << handle << " is invalid or not in use";
    errno = EBADF;
    return;
  }
  vpp = cvpp;
}

			// Lookup a vnode with a given cache file handle.
			// Return VListPtr describing the result. 
			// If the handle was invalid OR that vnode
			// is invalid or not in use, return
			// an invalid VListPtr and set errno to EBADF
VNode_list::VListPtr::VListPtr(const VNode_list& list, const int cache_handle)
	: vpp(0)
{
  if( cache_handle <= 0 )
  {
    Logger() << "Cache handle " << cache_handle << " is invalid";
    errno = EBADF;
    return;
  }
  register VNode **cvpp = list.vnode_array;
  register int i;
  for(i=1, cvpp=list.vnode_array; i<=list.size; i++,cvpp++)
  {
    if( *cvpp == 0 || !(bool)(**cvpp)  )
      continue;
    if( (**cvpp).get_cache_handle() == cache_handle )
    {
      handle = i; vpp = cvpp;
      return;
    }
  }
}

			// Create or lookup a vnode with a given name
			// If a vnode with that name is cached, return
			// its handle. Otherwise, create a new vnode
			// Return its handle (a positive number);
			// If failed, return an invalid handle
			//  and set errno appropriately
VHandle VNode_list::open(const char name [], const int open_flags)
{
  Logger clog; 
  clog << "Opening a v-node " << name << " with flags " << open_flags << endl;
  VListPtr vptr(*this,name);
  if( (bool)vptr )
  {
    clog << " found a cached " << vptr.handle
	 << ((bool)(**(vptr.vpp)) ? ", valid" : ",invalid");
    (**(vptr.vpp)).open(open_flags);
    (**(vptr.vpp)).assert_consistency();
    return vptr.handle;
  }
  if( !(bool)vptr.handle )
    vptr = VListPtr(*this,VListPtr::PURGE_GET_FREE());
  if( !(bool)vptr.handle )
    return clog << " no more free vnodes ", errno=ENFILE, vptr.handle;
  assert( vptr.vpp != 0 && *(vptr.vpp) == 0 );
  *(vptr.vpp) = 
  	name[strlen(name)-1] == '/' ? new VNodeDir(name) :
   		new VNodeFile(name);
  (**(vptr.vpp)).open(open_flags);
  (**(vptr.vpp)).assert_consistency();
  return clog << "allocated a new " << vptr.handle, 
     vptr.handle;
}

			// Locate a Vnode based on its VHandle and close it
			// Return false if that Vnode can't be located or
			// it wasn't open (errno is set appropriately then)
bool VNode_list::close(const VHandle handle)
{
  Logger() << "Closing " << handle;
  VListPtr vptr(*this,handle);
  if( !(bool)vptr )
    return false;
  return (**(vptr.vpp)).close();
}

			// Locate a Vnode based on a filehandle of
			// a (presumably open) cache file, and close it
			// -- both the cache file and the vnode
			// Return false if that Vnode can't be located or
			// it wasn't open (errno is set appropriately then)
bool VNode_list::close_cache_handle(const int cache_file_handle)
{
  Logger() << "Closing cache file handle" << cache_file_handle;
  VListPtr vptr(*this,cache_file_handle);
  if( !(bool)vptr )
    return false;
  return (**(vptr.vpp)).close();
}

			// Locate a Vnode based on its VHandle and return its
			// cache file. Return "" if the VNode couldn't be located
			// was invalid or wasn't open
const char * VNode_list::get_cache_file(const VHandle handle) const
{
  VListPtr vptr(*this,handle);
  if( !(bool)vptr )
    return "";
  return (**(vptr.vpp)).get_cache_file();
}


			// Locate a Vnode based on its VHandle and return the
			// handle of its opened cache file. Return -1 
			// if the VNode couldn't be located
			// was invalid or wasn't open
int VNode_list::get_cache_handle(const VHandle handle) const
{
  VListPtr vptr(*this,handle);
  if( !(bool)vptr )
    return -1;
  return (**(vptr.vpp)).get_cache_handle();
}


		// Send the extra headers and the request body
		// 'if_modified_since_time', if non-zero, tells us
		// to ask the server for updates, that is, what have
		// been changed since if_modified_since_time EPOCH time
void RequestContent_IfModif::send_request_data (ostream& link)
{ 
  if( modif_time > 0 )
  {
    char buffer[60]; buffer[0] = '\0';
    strftime(buffer,sizeof(buffer)-1,"%a, %d %b %Y %X GMT",
    	     gmtime(&modif_time));
    Logger() << "If-Modified-Since: " << buffer;
    link << "If-Modified-Since: " << buffer << CRLF;
  }
  RequestContent::send_request_data(link);
}

			// Try to obtain the status of a given file
			// or directory, if it is available in cache.
			// It doesn't matter if the vnode opened or not
			// As we don't know if the given name refers to
			// a directory, we look up both ways, with and
			// without the trailing slash.
FileStatus VNode_list::get_stat(const char name [])
{
  Logger clog; 
  clog << "Trying to get stat of a v-node " << name << endl;
  VListPtr vptr(*this,name);
  if( !(bool)vptr && name[strlen(name)-1] != '/' )
  {
    clog << "Repeating the lookup with / appended ";
    char buffer[VPATH_MAX+1];
    xstrncpy(buffer,name,sizeof(buffer)-3);
    strcat(buffer,"/");
    vptr = VListPtr(*this,buffer);
  }
  if( !(bool)vptr )
    return clog << "Lookup failed", FileStatus();
  const VNode & vnode = **(vptr.vpp);
  vnode.assert_consistency();
  clog << "Found cached " << vptr.handle;
  if( !(bool)vnode )
    return clog << "Found vnode was invalid",
    FileStatus();
    		// Probably we should re-validate the cache entry
    		// if more than say 1 hr after last_checked and
    		// nobody is using this vnode
  return vnode.get_stat();
}
