/*
 *----------------------------------------------------------------------------
 *			MCFS - HTTPFS Adapter
 *
 * This program is a server for a Midnight Commander's MCFS File System,
 * and a client for the HTTPFS. This code plumbs the two file systems
 * together, translating one file/directory access protocol
 * into the other.
 * 
 * The present source code file implements a MC server part of the adapter,
 * see mchttp-client.cc for a HTTPFS client part. The latter file also
 * explicitly spells out translation rules from MCFS to HTTPFS.
 * 
 * This code is based on the mcserv.c code by:
 * <BLOCKQUOTE>
   Copyright (C) 1995, 1996, 1997 The Free Software Foundation

   Written by: 
      Miguel de Icaza, 1995, 1997, 
      Andrej Borsenkow 1996.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 * </BLOCKQUOTE>
 *
 *
 * $Id: mcserv.cc,v 3.3 1999/05/21 18:23:40 oleg Exp oleg $
 *----------------------------------------------------------------------------
 */

/* {{{ Includes and global variables */

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <fcntl.h>
#include <string.h>
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <utime.h>
#include <assert.h>

/* Network include files */
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>

extern "C" {
#if 0
#include "../src/fs.h"
#else
#include "fs.h"
#endif
#include "mcfs.h"
#include "tcputil.h"
}

#include "mchttp-client.h"

#include "myenv.h"
#include "Logger.h"
#include <fstream.h>


/* The socket from which we accept commands */
static int msock;

/* Requested version number from client */
static int clnt_version;

/* Home directory */
static char *home_dir = NULL;

static char *up_dir = NULL;

/* Were we started from inetd? */
static int  inetd_started = 0;

/* Are we running as a daemon? */
static int  isDaemon = 0;

static char Curr_directory [512];

/* port number in which we listen to connections,
 * if zero, we try to contact the portmapper to get a port, and
 * if it's not possible, then we use a hardcoded value
 */
static int portnum = 0;


static int debug = 1;
static int quit_server;

/* }}} */


static VNode_list VNodes(16);


/* {{{ Misc routines */



void send_status (int status, int errno_number)
{
    rpc_send (msock, RPC_INT, status, RPC_INT, errno_number, RPC_END);
    errno = 0;
}

/* }}} */

/* {{{ File with handle operations */

void do_open (void)
{
  int flags, mode;
  char *arg;

  rpc_get (msock, RPC_STRING, &arg, RPC_INT, &flags, RPC_INT, &mode,RPC_END);
  Logger() << "Request: open: '" << arg << "', flags " << flags
           << ", mode " << mode;
  // Note! open flags except O_RDWR, O_RDONLY and O_WRONLY are not
  // portable! For example, O_CREAT os 0100 on Linux386 and 0000400 on HP-PA.
  // Therefore, we forcibly set O_CREAT for any file opened for writing.
  if( (flags & O_ACCMODE) != O_RDONLY )
    flags |= O_CREAT;
  VHandle handle = VNodes.open(arg,flags);
  free(arg);
  if( !(bool)handle )
  {
    Logger() << "V-node opening failed";
    send_status(-1,errno);
    return;
  }
  const int cache_handle = VNodes.get_cache_handle(handle);
  send_status (cache_handle, errno);
#if 0
    handle = open (arg, flags, mode);
    send_status (handle, errno);
#endif
}

void do_read (void)
{
    int handle, count, n;
    void *data;
   
    rpc_get (msock, RPC_INT, &handle, RPC_INT, &count, RPC_END);
    data = malloc (count);
    if (!data){
	send_status (-1, ENOMEM);
	return;
    }
    Logger() << "count=" << count;
    n = read (handle, data, count);
    Logger() << "read result=" << n;
    if (n < 0){
	send_status (-1, errno);
	return;
    }
    send_status (n, 0);
    rpc_send (msock, RPC_BLOCK, n, data, RPC_END);
    
    free (data);
}

void do_write (void)
{
    int handle, count, status;
    char buffer [8192];
    
    rpc_get (msock, RPC_INT, &handle, RPC_INT, &count, RPC_END);
    status = 0;
    while (count > 0){
	int nbytes = count > sizeof(buffer) ? sizeof(buffer) : count;
	
	rpc_get (msock, RPC_BLOCK, nbytes, buffer, RPC_END);
	status = write (handle, buffer, nbytes);
	if( status == nbytes )
	  count -= nbytes;
        else
	{
          status = -1; break;
	}
    }
    send_status (status, errno);
}

void do_lseek (void)
{
    int handle, offset, whence, status;

    rpc_get (msock,
	     RPC_INT, &handle,
	     RPC_INT, &offset,
	     RPC_INT, &whence, RPC_END);
    status = lseek (handle, offset, whence);
    send_status (status, errno);
}

void do_close (void)
{
  int cache_handle;
    
  rpc_get (msock, RPC_INT, &cache_handle, RPC_END);
  if( VNodes.close_cache_handle(cache_handle) )
    send_status(0,0);
  else
    Logger() << "do_close failed", 
    send_status (-1, errno);
}

/* }}} */

/* {{{ Stat family routines */


static void send_time (int sock, time_t time)
{
    if (clnt_version == 1) {
	struct tm *tmp = localtime(&time);
	rpc_send (msock,
		  RPC_INT, tmp->tm_sec,  	/* sec */
		  RPC_INT, tmp->tm_min,         /* min */
		  RPC_INT, tmp->tm_hour,         /* hour */
		  RPC_INT, tmp->tm_mday,          /* mday */
		  RPC_INT, 1900+tmp->tm_year,         /* year */
		  RPC_INT, tmp->tm_mon,	         /* month */
		  RPC_END);
    } else {
	long ltime = (long) time;
	char buf[2*sizeof(long) + 1];

	sprintf (buf, "%lx", ltime);
	rpc_send (msock,
		  RPC_STRING, buf,
		  RPC_END);
    }
}

void send_stat_info (const struct stat& st)
{
    const int blocks =
#ifdef HAVE_ST_BLOCKS
	st.st_blocks;
#else
        st.st_size / 1024;
#endif
    
    rpc_send (msock, RPC_INT, (long) st.st_dev, 
	      RPC_INT, (long) st.st_ino,
	      RPC_INT, (long) st.st_mode,
	      RPC_INT, (long) st.st_nlink,
	      RPC_INT, (long) st.st_uid,
	      RPC_INT, (long) st.st_gid,
	      RPC_INT, (long) st.st_size,
	      RPC_INT, (long) blocks, RPC_END);
    send_time (msock, st.st_atime);
    send_time (msock, st.st_mtime);
    send_time (msock, st.st_ctime);
}

struct SendStatusInfo : public ContentConsumer
{
  bool operator () (istream& link, const RFC822Headers& headers)
  { FileStatus status_info(headers);
    if( !(bool)status_info )
      return (clog << "invalid stat"), false;
     send_status(0,0);
     send_stat_info(status_info);
     return true;
   }
   SendStatusInfo(void) {}
};


void do_lstat (void)
{
    char   *file = 0;

    rpc_get (msock, RPC_STRING, &file, RPC_END);
    Logger() << "Request: lstat: " << file;
    HTTPTransaction http_trans(
		VFSNameParser::vfs_name_to_httpfs(file));
    free (file);
    if( http_trans.perform_transaction(
	RequestWithCommand(RequestContent::HEAD,"lstat"),SendStatusInfo())
       != HTTPTransaction::OK )
    {
      Logger() << "Transaction error";
      send_status(-1,EIO);
      return;
    }
}

void do_fstat (void)
{
    int handle;
    int n;
    struct stat st;
   
    rpc_get (msock, RPC_INT, &handle, RPC_END);
    n = fstat (handle, &st);
    send_status (n, errno);
    if (n < 0)
	return;

    send_stat_info (st);
}

// Because of mc's quirly behavior (it tends to keep chopping
// of path components after chdir, which makes a VFS name invalid)
// we make sure that the VFS name contains at least one slash
// (not counting the leading slash)
void do_stat (void)
{
    char   *file = 0;

    rpc_get (msock, RPC_STRING, &file, RPC_END);
    Logger() << "Request: stat: " << file;
    if( file == 0 || file[0] == '\0' ||
	strchr(file+1,'/') == 0 )
    {
      Logger() << "Bad VFS file name";
      send_status(-1,EIO); 
      return;
    }
    FileStatus cached_status = VNodes.get_stat(file);
    if( (bool)cached_status )
    {
      Logger() << "sending cached status";
      send_status(0,0);
      send_stat_info(cached_status);
      free (file);
      return;
    }
    HTTPTransaction http_trans(
		VFSNameParser::vfs_name_to_httpfs(file));
    free (file);
    if( http_trans.perform_transaction(
	RequestWithCommand(RequestContent::HEAD,"stat"),SendStatusInfo())
       != HTTPTransaction::OK )
    {
      Logger() << "Transaction error";
      send_status(-1,EIO);
      return;
    }
}

/* }}} */

/* {{{ Directory lookup operations */


void do_opendir (void)
{
  char *arg;

  rpc_get (msock, RPC_STRING, &arg, RPC_END);

    		// make sure that a dir name ends in a slash
  char * dir_name = arg;
  if( dir_name[strlen(dir_name)-1] != '/' )
  {
    const int old_len = strlen(arg);
    dir_name = (char *)realloc(dir_name,old_len+2);
    assert( dir_name );
    dir_name[old_len] = '/';
    dir_name[old_len+1] = '\0';
  }
  VHandle handle = VNodes.open(dir_name,O_RDONLY);
  if( (bool)handle )
    send_status(handle.get_raw_handle(),0);
  else
    Logger() << "V-node opening failed",
    send_status(-1,errno == 0 ? EIO : errno);
  free(dir_name);
}

/* Sends the complete directory listing, as well as the stat information */
void do_readdir (void)
{
    int    handle_raw;
 
    rpc_get (msock, RPC_INT, &handle_raw, RPC_END);
    
    Logger clog;
    clog << "readdir for " << handle_raw;
    VHandle handle(handle_raw);
    const char * const cache_file = VNodes.get_cache_file(handle);
    clog << "Cache file for " << handle << " is >" 
	 << cache_file << "<";
    if( cache_file == 0 || *cache_file == '\0' )
    {
      rpc_send (msock, RPC_INT, 0, RPC_END);
      return;
    }
    ifstream dir_stream(cache_file);
    char buffer[100];
    if( !dir_stream.good() )
    {
      clog << "Failed to open the cache file " << cache_file;
      rpc_send (msock, RPC_INT, 0, RPC_END);
      return;
    }
    while( !dir_stream.eof() )
    {
      dir_stream.get(buffer,sizeof(buffer)-1,'/');
	  //clog << "read '" << buffer << "'" << endl;
      bool good_flag = dir_stream.get() == '/' || buffer[0] != '\0';
      if( !good_flag )
	xstrncpy(buffer,"io-error",sizeof(buffer)-1);
      rpc_send (msock, RPC_INT, strlen(buffer), RPC_END);
      rpc_send (msock, RPC_BLOCK, strlen(buffer), buffer, RPC_END);
      if( !good_flag )
      {
	clog << "invalid name"; send_status(-1, EIO);
	return;
      }
      dir_stream.getline(buffer,sizeof(buffer)-1,'\n');
      if( !dir_stream.good() )
      {
	send_status(-1, EIO);
	return;
      }
      FileStatus status(buffer);
	  //clog << "read stat: " << buffer << endl;
      if( !(bool)status )
      {
	clog << "invalid stat"; send_status(-1, EIO);
	return;
      }
      send_status(0,0);
      send_stat_info(status);
      dir_stream.peek();		// Forces EOF
    }
    rpc_send (msock, RPC_INT, 0, RPC_END);
}

void do_closedir (void)
{
  int raw_handle;

  rpc_get (msock, RPC_INT, &raw_handle, RPC_END);
  if( !VNodes.close(VHandle(raw_handle)) )
    Logger() << "error closing vnode #" << raw_handle;
}

/* }}} */

/* {{{ Operations with one and two file name argument */

void do_chdir (void)
{
    char *file;
    int  status;

    rpc_get (msock, RPC_STRING, &file, RPC_END);
    
#if 0
    status = chdir (file);
#else
    Logger() << "Request: chdir: " << file;
    if( strlen(file) < sizeof(Curr_directory) )
    {
      strcpy(Curr_directory,file);
      status = errno = 0;
    }
    else
      status = -1, errno = ENAMETOOLONG;
#endif
    send_status (status, errno);
    free (file);
}

void do_rmdir (void)
{
    char *file;
    int  status;

    rpc_get (msock, RPC_STRING, &file, RPC_END);
    
#if 0
    status = rmdir (file);
#else
    Logger() << "Request: rmdir: " << file;
    status = -1, errno = EIO;
#endif
    send_status (status, errno);
    free (file);
}

void do_mkdir (void)
{
    char *file;
    int  mode, status;

    rpc_get (msock, RPC_STRING, &file, RPC_INT, &mode, RPC_END);
#if 0
    status = mkdir (file, mode);
#else
    Logger() << "Request: mkdir: '" << file << "' mode " << mode;
    status = -1, errno = EIO;
#endif
    
    send_status (status, errno);
    free (file);
}

void do_mknod (void)
{
    char *file;
    int  mode, dev, status;

    rpc_get (msock, RPC_STRING, &file, RPC_INT, &mode, RPC_INT, &dev, RPC_END);
    
#if 0
    status = mknod (file, mode, dev);
#else
    Logger() << "Request: mknod: '" << file << "' mode " << mode
             << " dev " << dev;
    status = -1, errno = EACCES;
#endif
    send_status (status, errno);
    free (file);
}

void do_readlink (void)
{
    char *file;

    rpc_get (msock, RPC_STRING, &file, RPC_END);
    Logger() << "Request: readlink: " << file;
    HTTPTransaction http_trans(
		VFSNameParser::vfs_name_to_httpfs(file));
    free (file);
    
    struct SendSLinkTarget : public ContentConsumer
    {
      bool operator () (istream& link, const RFC822Headers& headers)
     { 
       char buffer [2048];
       link.get(buffer,sizeof(buffer)-1,'\n');
       if( link.get() != '\n' || buffer[0] == '\0' )
	 return false;
       send_status(strlen(buffer),0);
       rpc_send (msock, RPC_STRING, buffer, RPC_END);
       return true;
     }
     SendSLinkTarget(void) {}
    };

    if( http_trans.perform_transaction(
	RequestWithCommand(RequestContent::GET,"readlink"),SendSLinkTarget())
       != HTTPTransaction::OK )
    {
      Logger() << "Transaction error";
      send_status(-1,EIO);
      return;
    }
#if 0
    n = readlink (file, buffer, 2048);
    send_status (n, errno);
    if (n >= 0) {
        buffer [n] = 0;
	rpc_send (msock, RPC_STRING, buffer, RPC_END);
    }
    free (file);
#endif
}

void do_unlink (void)
{
    char *file;
    int  status;
    
    rpc_get (msock, RPC_STRING, &file, RPC_END);
#if 0
    status = unlink (file);
#else
    Logger() << "Request: unlink: " << file;
    status = -1, errno = EIO;
#endif
    send_status (status, errno);
    free (file);
}

void do_rename (void)
{
    char *f1, *f2;
    int  status;
    
    rpc_get (msock, RPC_STRING, &f1, RPC_STRING, &f2, RPC_END);
#if 0
    status = rename (f1, f2);
#else
    Logger() << "Request: rename: '" << f1 << "' to '" << f2 << "'";
    status = -1, errno = EIO;
#endif
    send_status (status, errno);
    free (f1); free (f2);
}

void do_symlink (void)
{
    char *f1, *f2;
    int  status;
    
    rpc_get (msock, RPC_STRING, &f1, RPC_STRING, &f2, RPC_END);
#if 0
    status = symlink (f1, f2);
#else
    Logger() << "Request: symlink: '" << f1 << "' to '" << f2 << "'";
    status = -1, errno = EACCES;
#endif
    send_status (status, errno);
    free (f1); free (f2);
}

void do_link (void)
{
    char *f1, *f2;
    int  status;
    
    rpc_get (msock, RPC_STRING, &f1, RPC_STRING, &f2, RPC_END);
#if 0
    status = link (f1, f2);
#else
    Logger() << "Request: link: '" << f1 << "' to '" << f2 << "'";
    status = -1, errno = EACCES;
#endif
    send_status (link (f1, f2), errno);
    free (f1); free (f2);
}


/* }}} */

/* {{{ Misc commands */

void do_gethome (void)
{
    rpc_send (msock, RPC_STRING, (home_dir) ? home_dir : "/", RPC_END);
}

void do_getupdir (void)
{
    rpc_send (msock, RPC_STRING, (up_dir) ? up_dir : "/", RPC_END);
}

void do_chmod (void)
{
    char *file;
    int  mode, status;
    
    rpc_get (msock, RPC_STRING, &file, RPC_INT, &mode, RPC_END);
#if 0
    status = chmod (file, mode);
#else
    Logger() << "Request: chmod: '" << file << "' mode " << mode;
    Logger() << "Replying OK but not actually changing the mode...";
//    status = -1, errno = EPERM;
    status = 0, errno = 0;
#endif
    send_status (status, errno);
    free (file);
}

void do_chown (void)
{
    char *file;
    int  owner, group, status;
    
    rpc_get (msock, RPC_STRING, &file,RPC_INT, &owner, RPC_INT,&group,RPC_END);
#if 0
    status = chown (file, owner, group);
#else
    Logger() << "Request: chown: '" << file << "' owner " << owner
    	     << " group " << group;
    status = -1, errno = EPERM;
#endif
    send_status (status, errno);
    free (file);
}

void do_utime (void)
{
    char *file;
    int  status;
    long atime;
    long mtime;
    char *as;
    char *ms;
    struct utimbuf times;
    
    rpc_get (msock, RPC_STRING, &file,
		    RPC_STRING, &as,
		    RPC_STRING, &ms,
		    RPC_END);
    sscanf (as, "%lx", &atime);
    sscanf (ms, "%lx", &mtime);
    Logger() <<  "Got a = " << as << ", m = " << ms
	     << ", comp a = " << atime << ", m = " << mtime;
    free (as);
    free (ms);
    times.actime  = (time_t) atime;
    times.modtime = (time_t) mtime;
#if 0
    status = utime (file, &times);
#else
    Logger() << "Request: utime: '" << file;
    status = -1, errno = EPERM;
#endif
    send_status (status, errno);
    free (file);
}

void do_quit ()
{
    quit_server = 1;
}

/* This implements nothing _but_ the anonymous login */
static void do_login ()
{
    char *username;
    int  result;
    
    rpc_get (msock, RPC_LIMITED_STRING, &up_dir, RPC_LIMITED_STRING, &username, RPC_END);
    Logger() << "Login: username: " << username;
    bool logged_in = true; // log in is always successful, at present
    rpc_send (msock, RPC_INT,
	      logged_in ? MC_LOGINOK : MC_INVALID_PASS,
	      RPC_END);
}

/* }}} */

/* {{{ Server and dispatching functions */

/* This structure must be kept in synch with mcfs.h enums */

static struct _command {
    char *command;
    void (*callback)(void);
} commands [] = {
    { "open",       do_open },
    { "close",      do_close },
    { "read",       do_read },
    { "write",      do_write },
    { "opendir",    do_opendir }, 
    { "readdir",    do_readdir },
    { "closedir",   do_closedir },
    { "stat ",      do_stat },
    { "lstat ",     do_lstat },
    { "fstat",      do_fstat },
    { "chmod",      do_chmod },
    { "chown",      do_chown },
    { "readlink ",  do_readlink },
    { "unlink",     do_unlink },
    { "rename",     do_rename },
    { "chdir ",     do_chdir },
    { "lseek",      do_lseek },
    { "rmdir",      do_rmdir },
    { "symlink",    do_symlink },
    { "mknod",      do_mknod },
    { "mkdir",      do_mkdir },
    { "link",       do_link },
    { "gethome",    do_gethome },
    { "getupdir",   do_getupdir },
    { "login",      do_login },
    { "quit",       do_quit },
    { "utime",      do_utime },
};

void exec_command (const int command)
{
    if (command < 0 ||
	command >= sizeof(commands)/sizeof(commands[0]) ||
	commands [command].command == 0){
	Logger() << "Got unknown command: " << command;
	exit(4);
    }
    Logger() << "Command: " << commands[command].command;
    (*commands [command].callback)();
}

static void check_version (void)
{
    int version;
    
    rpc_get (msock, RPC_INT, &version, RPC_END);
    if (version >= 1 &&
	version <= RPC_PROGVER)
	rpc_send (msock, RPC_INT, MC_VERSION_OK, RPC_END);
    else
	rpc_send (msock, RPC_INT, MC_VERSION_MISMATCH, RPC_END);
    
    clnt_version = version;
}

/* This routine is called by rpc_get/rpc_send when the connection is closed */
extern "C" {
void tcp_invalidate_socket (int sock);
};  
void tcp_invalidate_socket (int sock)
{
    fprintf (stderr,"\nConnection broken\n");
    exit(1);
}

static /*volatile*/ void serve_new_connection (const int sock)
{
    int command;
    
    msock = sock;
    quit_server = 0;

    check_version ();
    do {
	if( rpc_get (sock, RPC_INT, &command, RPC_END) )
			exec_command (command);
    } while (!quit_server);
}

/* }}} */

/* {{{ Net support code */

static /*volatile*/ void serve_client_connections (const int portnum)
{
    int sock, clilen, newsocket;
    struct sockaddr_in client_address, server_address;
    struct hostent *hp;
    char hostname [255];
    int yes = 1;
    
    if ((sock = socket (AF_INET, SOCK_STREAM, 0)) < 0)
	perror("Can't create socket"), exit(4);

    /* Use this to debug: */
    if (setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, (char *) &yes, sizeof (yes)) < 0)
	perror("setsockopt failed"), exit(4);

    gethostname (hostname, 255);
    Logger () <<  "hostname=" << hostname;
    hp = gethostbyname (hostname);
    if(hp == 0)
      perror("hp = 0!"), exit(4);
    
    memset((char *) &server_address, 0, sizeof (server_address));
    server_address.sin_family = hp->h_addrtype;
    server_address.sin_addr.s_addr = htonl (INADDR_ANY);
    server_address.sin_port = htons (portnum);

    if (bind (sock, (struct sockaddr *) &server_address,
	      sizeof (server_address)) < 0)
	perror("Can't bind"), exit(4);

    listen (sock, 5);

    for (;;){
	int child;
	
	clilen = sizeof (client_address);
	newsocket = accept (sock, (struct sockaddr *) &client_address,
			    &clilen);

	if (isDaemon && (child = fork())) {
	    int status;
	    
	    close (newsocket);
	    waitpid (child, &status, 0);
	    continue;
	}

	serve_new_connection(newsocket); // does not return
	//close (newsocket);
    }
}



/* }}} */

int main (int argc, char *argv [])
{
    char *result;
    extern char *optarg;
    int verbose = 0;
    
    do
      switch (getopt (argc, argv, "diqp:v"))
      {
	case -1:
	    break;
	    
	case 'd':
	    isDaemon = 1;
	    verbose = 0;
	    continue;

	case 'v':
	    verbose = 1;
	    continue;
	    
	case 'q':
	    verbose = 0;
	    continue;

	case 'p':
	    portnum = atoi (optarg);
	    continue;

	case 'i':
	    inetd_started = 1;
	    continue;

	default:
	    fprintf (stderr, "Usage is: mcserv [options] [-p portnum]\n\n"
		    "options are:\n"
		    "-d  become a daemon (sets -q)\n"
		    "-q  quiet mode\n"
		    "-v  verbose mode\n"
		    "-p  to specify a port number to listen\n");
	    exit (0);
    } while(0);

    if (isDaemon && fork()) exit (0);

    if (portnum == 0)
	portnum = get_port_number ();

    if (portnum == -1)
      Logger() << "Failed to get the port number!" << endl, exit(4);
  
    register_port (portnum);
    Logger() << "Using port " << portnum << endl;
#ifdef HAVE_PMAP_SET
	if (!isDaemon) // do it via atexit()
	    pmap_unset (RPC_PROGNUM, RPC_PROGVER);
#endif
    serve_client_connections(portnum); // never returns!
}
