// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *			HTTP transport layer
 *
 * 	Standard utility services used by many parts of the project
 *
 * $Id: util.cc,v 2.2 1998/04/20 18:46:01 oleg Exp $
 *
 ************************************************************************
 */

#ifdef __GNUC__
#pragma implementation
#endif

#include <iostream.h>
#include "util.h"
#include <time.h>
#include <ctype.h>


//------------------------------------------------------------------------
//		Handling of a collection of strings
//
// Individual strings' (null-terminated) bodies are stuffed into a into a buffer.
// This class serves as a string 'allocator'
// To add a new string, one calls a new_string() method, and note the mark
// it returns.

			// Dump the current status of the watchdog
ostream& RWWatchDog::dump(ostream& os)
{
  if( !q_engaged() )
    os << "The watchdog object is not engaged";
  else if( q_exclusive() )
    os << "The watchdog object is being exclusively locked";
  else
    os << "The watchdog object is shared among " << ref_count << " readers";
  return os << endl;
}

const StringCollection::Mark StringCollection::Mark::invalid_mark((size_t)EOF);

			// Start the new string and return a mark
			// where it begins
StringCollection::Mark StringCollection::new_string(void)
{
  get_exclusive();
  return filled_size;
}

			// Add a new character to the collection,
			// which must be in the writing (exclusive)
			// mode
			// We use a Fibonacci expansion when we
			// need to reallocate the current buffer
StringCollection& StringCollection::operator << (const char c)
{
  assert( q_exclusive() && filled_size <= current_bsize );
  if( filled_size == current_bsize )
  {
    const size_t new_size = current_bsize + prev_bsize;
    prev_bsize = current_bsize;
    assert( (buffer = (char *)realloc(buffer,current_bsize = new_size)) );
  }
  buffer[filled_size++] = c;
  return *this;
}

			// Add a few characters to the collection
StringCollection& StringCollection::operator << (const char * str)
{
  while( *str != '\0' )
    *this << *str++;
  return *this;
}

			// Terminate the string and
			// switch the writing mode off
StringCollection& StringCollection::operator << (void (*dummy)(Ends))
{
  *this << '\0';
  release_exclusive();
  return *this;
}

			// Writing out the contents of the mark
			// (used mostly for debugging purposes)
ostream& operator << (ostream& os, const StringCollection::Mark& mark)
{
  if( !mark )
     os << "Invalid mark";
  else
     os << mark.pos;
  return os;
}

			// Dump the whole collection
  			// (for debugging purposes mostly)
ostream& StringCollection::dump(ostream& os)
{
  os << "***String collection: " << endl;
  os << "Currently allocated " << current_bsize
     << " bytes; previous allocation " << prev_bsize << endl;
  os << "Currently filled " << filled_size << endl;
  RWWatchDog::dump(os);
  for(const char * p = buffer; p < buffer+filled_size;)
  {
    os << "Pos " << (p-buffer) << ": >";
    while( p < buffer+filled_size )
      if( const char c = *p++ )
        os << c;
      else
        break;			// if \0 encountered, scan past it and exit
    os << "<" << endl;	
  }
  return os << "--- end-of-collection---\n" << endl;
}


//------------------------------------------------------------------------
//		Handling of a Light-Weight input string stream

			// Construct a iststream to read a particular
			// string from a particular collection
			// Note, the mark may be invalid. In that case,
			// the iststream is marked _bad_
void iststream::init(const StringCollection::Mark mark)
{
  if( !mark )
  {
    setstate(ios::badbit);
    return;
  }
  
  curr_ptr = q_mark_ptr(mark);
  end_ptr = q_end_buffer();
}

iststream::iststream(const StringCollection& collection,
		     const StringCollection::Mark mark)
#ifndef __BORLANDC__
  : StringCollection::CollectionRef(collection),
#else
  : CollectionRef(collection),
#endif
    curr_ptr(0), end_ptr(0), state(0)
{
  init(mark);
}

  			// Lookup a value with a given key in the dictionary,
  			// and build the iststream to access it
  			// The iststream may be bad if the key wasn't found...
iststream::iststream(const Dictionary& dictionary, const char key[])
#ifndef __BORLANDC__
  : StringCollection::CollectionRef(dictionary.buffer),
#else
  : CollectionRef(dictionary.buffer),
#endif
    curr_ptr(0), end_ptr(0), state(0)
{
  init(dictionary.lookup(key));
}

iststream::iststream(const Dictionary& dictionary,
		     const StringCollection::Mark mark)
#ifndef __BORLANDC__
  : StringCollection::CollectionRef(dictionary.buffer),
#else
  : CollectionRef(dictionary.buffer),
#endif
    curr_ptr(0), end_ptr(0), state(0)
{
  init(mark);
}

			// Get the next character from the current string's
			// body
			// EOF and '\0' are both treated as the
			// string's end (and raise the EOF condition and
			// cause EOF returned)
int iststream::get(void)
{
  assert( good() );
  if( curr_ptr < end_ptr && *curr_ptr != '\0' )
    return *curr_ptr++;
  setstate(ios::eofbit);
  return EOF;
}

			// Same as get() above; only the returned symbol
			// is left "unread" and can be get() next time around
			// Peek-ing EOF raises the EOF condition on the
			// stream!
int iststream::peek(void)
{
  if( eof() )
    return EOF;
  assert( good() );
  if( curr_ptr < end_ptr && *curr_ptr != '\0' )
    return *curr_ptr;
  setstate(ios::eofbit);
  return EOF;
}

			// Put a char 'c' back into the iststream
			// We assume that this was the character
			// which was last read from the stream
iststream& iststream::putback(const char c)
{
  assert( good() );
  assert( *--curr_ptr == c );
  return *this;
}

			// Dump the current string to ostream
ostream& operator << (ostream& os, iststream& sa)
{
  register int c;
  while( (c = sa.get()) != EOF )
    os << (char)c;
  return os;
}

			// Load a string dest_str (of size at most n) from this 
			// string stream (StringAccessor&)
			// until EOF or one of the terminators encountered
			// The terminator is left on the stream
iststream&
iststream::getline_upto(char * dest_str, const int n,
    			const char terminators [])
{
  const char * const str_end = dest_str + n - 1;
  for(register char c = peek(); c != EOF; c = (get(), peek()))
    if( strchr(terminators,c) )
      break;
    else if( dest_str >= str_end )
      _error("string too long");
    else
      *dest_str++ = c;
  *dest_str = '\0';
  return *this;
}


    		// The same as above, but reads and discards the delimiter
iststream& iststream::getline(char* ptr, int len, char delim)
{
  const char terminators [] = {delim,0};
  getline_upto(ptr,len,terminators);
  if( !eof() )
    assert( get() == delim );
  return *this;
}

		// Read an integral value out of the istream
iststream& iststream::operator >> (int& val)
{
  bool negative = false;
  enum { Start, Seen_sign, Accumulating } status = Start;
  int accum = 0;
  
  for(register int c = peek(); c != EOF; c=(get(),peek()))
  {
    switch(status)
    {
      case Start:
      case Seen_sign:
           if( isspace(c) )
             continue;
           else if( c == '+' )
             if( status == Seen_sign)
               break;			// second sign is an error
             else
               { status = Seen_sign; continue;}
           else if( c == '-' )
             if( status == Seen_sign)
               break;			// second sign is an error
             else
               {negative=true; status = Seen_sign; continue;}
           else if( !isdigit(c) )
             break;	// The first significant char must be a digit...
           status = Accumulating;
    	
               			// Falling through: a first digit is
               			// spotted
      case Accumulating:
           if( !isdigit(c) )
             break;		// a non-digit terminates it
            accum = accum*10 + (c-'0');
            continue;
            
      default:
           assert(0);		// Can't happen
    }
    break;
  }
  if( status != Accumulating )
    setstate(ios::failbit);
  else
    val = negative ? -accum : accum;
  return *this;
}

  		// Tell the current position in the stream
  		// Return the invalid mark if the stream is not in a good
  		// state
StringCollection::Mark iststream::tellg(void) const
{
  if( !good() )
    return StringCollection::Mark::invalid_mark;
  return q_ptr_mark(curr_ptr);
}

                                // Convert char c to lower case
                                // Unlike traditional tolower(), it
                                // applies conversion only if c is a letter
                                // in uppercase
                                // And it is not a macro!
static inline int to_lower(const char c)
{
  return isupper(c) ? (c - 'A') + 'a' : c;
}


			// Case-insensitive check if the iststream contains
			// 'str'(as a prefix), starting from its current
			// position.
			// If the stream does contain the entire str, it is
			// positioned to read the first character after the
			// str is scanned.
			// Otherwise, the EOF condition is set on the stream
iststream& iststream::begins_with_ci(const char * str)
{
  while( *str != '\0' )
  {
    register const int c = get();
    if( c == EOF )
      break;				// the stream now will be in the EOF state
    else if( to_lower(c) != to_lower(*str++) )
    {
      setstate(ios::eofbit);
      break;
    }
  }
  return *this;
}

			// Case-insensitive check to see if the
			// iststream contains 'str', and nothing more
bool strcasecmp(const char * str, iststream& sa)
{
  register int c;
  while( (c=sa.get()) != EOF )
    if( to_lower(c) != to_lower(*str++) )
      return false;
  return *str == '\0';		// sa is finished, check is str is over too
}

			// Case-insensitive comparison of two iststreams...
bool strcasecmp(iststream& s1, iststream& s2)
{
  register int c1, c2;
  for(;;)
  {
    if( (c1=s1.get()) == (c2=s2.get()) )
      if( c1 == EOF )
        return true;
      else
        continue;
    if( to_lower(c1) != to_lower(c2) )
      break;
  }
  return false;
}


//------------------------------------------------------------------------
//  A Dictionary of key-value string pairs, with a predefined search path
//
// The items added last would be looked at first during lookup.
// The bodies of the terminated key and value strings are stored in the
// StringCollection buffer; a list of the associations contains only marks
// in the buffer

				// Add a new association to the dictionary
				// (the bodies of the key and the value are
				// supposed to be in the buffer already).
void Dictionary::add_assoc(const StringCollection::Mark pos_key,
			   const StringCollection::Mark pos_value)
{
  const int free_slot = no_assoc++;
  if( free_slot >= no_assoc_slots )	// Reallocation is necessary
    if( associations != assoc_buffer )
       associations = (Assoc *)
       		      realloc(associations,(no_assoc_slots *= 2)*sizeof(Assoc));
     else
     {
     	assert( associations = (Assoc *)calloc(no_assoc_slots *= 2,sizeof(Assoc)) );
     	memcpy(associations,assoc_buffer,sizeof(assoc_buffer));
     }
  assert( associations != 0 && free_slot < no_assoc_slots );
  new(associations+free_slot) Assoc(pos_key,pos_value);
}



			// Dump the contents of the associations
ostream& Dictionary::Assoc::dump(ostream& os, StringCollection& buffer) const
{
  return os << iststream(buffer,key) << ": " << iststream(buffer,value);
}

ostream& operator << (ostream& os, const Dictionary& dict)
{
  if( dict.no_assoc == 0 )
    return os << "\n\t[Empty]" << endl;

  for(register const Dictionary::Assoc * ap = dict.associations+dict.no_assoc;
      --ap >= dict.associations; )
  {
     os << "\t"; ap->dump(os,dict.buffer) << endl;
  }
  return os;
}


			// Looking for a specific association
			// with a given 'key'

  				// Lookup a value with a given key and
  				// return buffer's position of the value
  				// Return an invalid mark if not found
StringCollection::Mark Dictionary::lookup(const char key[]) const
{
  for(register const Assoc * ap = associations+no_assoc; --ap >= associations;)
    if( strcasecmp(key,iststream(buffer,ap->key)) )
      return ap->value;
  return StringCollection::Mark::invalid_mark;
}


			// Look up an integral value associated with a given
			// key
			// Return the found value, or fallback's value if
			// the key wasn't found, or its value cannot be
			// converted to int
int Dictionary::lookup(const char key[], const int fallback) const
{
  const StringCollection::Mark pos_value = lookup(key);
  if( !pos_value  )
    return fallback;
  int value = fallback;
  if( (iststream(buffer,pos_value) >> value).fail() )
    return fallback;
  return value;
}

			// Look up a value associated with a key
			// pass it over to a value_taker (and return
			// its value)
			// Return false if the key wasn't found
bool Dictionary::lookup(const char key[],
			const DictValueHandler& value_taker) const
{
  const StringCollection::Mark pos_value = lookup(key);
  return (bool)pos_value && value_taker(iststream(buffer,pos_value));
}

//........................................................................
// 		Iterate through the dictionary
// For each dictionary association (starting with the last one), we evaluate
//   key_value_handler.handle_key(current_assoc_key)
// If it returns true, we evaluate
//   key_value_handler.handle_value(current_assoc_value)
// If this returns true, we stop and return true;  Otherwise, we move to
// the next association in the dictionary. If we've gone through all the
// associations, we return false
//
// In other words, this predicate tries to ascertain an assertion
// There_Exists[ assoc(a) : key_value_handler.handle_key(a.key) &&
//			    key_value_handler.handle_value(a.value) ]
//
// Usage examples:
//  - traversal through all keys: handle_key() to always return false
//  - traversal through all keys and values:
//	handle_key() to always return true,
//	handle_value() to always return false
//  - looking up an element with some key:
//	handle_key() checks to see if the key is the one we look for;
//	handle_value() to return true (to terminate the search)

bool Dictionary::any(const DictKeyValueHandler& key_value_handler) const
{
  for(register const Assoc * ap = associations+no_assoc; --ap >= associations;)
    if( key_value_handler.handle_key(iststream(buffer,ap->key)) &&
        key_value_handler.handle_value(iststream(buffer,ap->value)) )
      return true;
  return false;
}
