// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *			   HTTP transport layer
 *
 * 	Standard utility services used by many parts of the project
 *
 * A bag of strings, a dictionary, and light-weight streams. See comments
 * below for more details. See vutil.cc for sample code that uses
 * these facilities in common situations, and makes sure they
 * really work.
 *
 * $Id: util.h,v 2.3 1998/07/30 16:08:13 oleg Exp $
 *
 ************************************************************************
 */

#ifndef __GNUC__
#pragma once
#else
#pragma interface
#endif

#ifndef __util_h
#define __util_h

#include "myenv.h"
#include "std.h"
#include "Logger.h"
#include <iostream.h>

class Dictionary;

		// The following class watches over exclusive (write)
		// and shared (read) access to a data structure the object
		// is a member of.
		// The ref_count field tells how the object is being
		// currently accessed:
		//	= 0: the object is not engaged
		//	=-1: exclusive (write access)
		//	>0 : the object is being currently read (viewed)
		//	by one or several readers
		// The watchdog watches over obvious blunders as trying
		// to get a read access while the object is exclusively
		// held, or trying to grab the write access or destroy the
		// object while it is engaged in some other activity.
class RWWatchDog
{
  int ref_count;			// see explanation above

  RWWatchDog(const RWWatchDog&);	// Deliberately unimplemented:
  void operator = (const RWWatchDog&);	// no cloning allowed!

public:
  bool q_engaged(void) const		{ return ref_count != 0; }
  bool q_shared(void) const		{ return ref_count > 0; }
  bool q_exclusive(void) const		{ return ref_count == -1; }
  
  void get_exclusive(void)
  	{ assert(!q_engaged()); ref_count = -1; }
  void release_exclusive(void)
  	{ assert(q_exclusive()); ref_count = 0; }
  	
  void get_shared(void)
  	{ assert(!q_exclusive()); ref_count++; }
  void release_shared(void)
  	{ assert(q_shared()); ref_count--; }
  	
  RWWatchDog(void) : ref_count(0) {}
  ~RWWatchDog(void) { assert(ref_count==0); ref_count = -1; }
  
  ostream& dump(ostream& os);	// Dump the current status
};

		// A collection of strings
		// Strings' (null-terminated) bodies are all
		// stuffed into a buffer.
		// This class serves as a string 'allocator'
		// To add a new string, one calls a
		// new_string() method, and note the mark it returns.
		// The new_string() methods turns the buffer into the
		// exclusive (writing) mode. While the buffer is in this
		// mode, one can add new characters (to the end of buffer)
		// via an operator <<.
		// *this << ends; finishes the string and releases
		// the buffer.
		// The string can then be accessed by constructing an
		// iststream and passing the mark of the desired string
		// (the mark noted when the string was "opened")

class StringCollection : private RWWatchDog
{
  class CollectionRef;		// Borland needs a forward decl of a nested class
  friend class CollectionRef;

  char * buffer;		// Where the strings bodies lie
  				// May be reallocated when new chars/strings
  				// are added!
  enum { init_allocation = 100 };
  size_t current_bsize;		// Current size of the buffer
  size_t prev_bsize;		// size of the buffer before it was
  				// reallocated
  size_t filled_size;		// Length of the currently filled part of
  				// the buffer

  StringCollection(const StringCollection&);	// Deliberately unimplemented:
  void operator = (const StringCollection&);	// no cloning allowed!

public:
  StringCollection(void) :
	buffer((char *)malloc(init_allocation)),
	current_bsize(init_allocation),
  	prev_bsize(init_allocation),
  	filled_size(0)
  	{ assert( buffer ); }
  	
			// An opaque class that denotes a specific
			// location in a collection of strings
			// (usually the beginning of a string)
			// A mark may also denote no location at all:
			// when (bool)Mark returns false.
			// The mark keeps an offset in the collection
			// The reason we use the offset rather than a
			// pointer is that the collection's buffer
			// may be reallocated (thus invalidating the
			// pointers)
  class Mark			
  {
    friend class StringCollection;
    friend class CollectionRef;
    friend class iststream;
    friend class Dictionary;
    size_t pos;
    Mark(void) {}				// private constructor
    Mark(size_t _pos) : pos(_pos) {}		// private constructor
  public:
    operator bool (void) const		{ return pos != (size_t)EOF; }
    bool operator ! (void) const	{ return pos == (size_t)EOF; }
    static const Mark invalid_mark;

    class DONT_CONSTRUCT {};
    Mark(DONT_CONSTRUCT) {}
    friend ostream& operator << (ostream& os, const Mark& mark);
  };
  
  Mark new_string(void);	// Switch the Collection into the
  				// writing mode and return the mark
  				// of thus started string
  				
  				// While the buffer is in this mode, one
  				// can use the following methods:
  			
  StringCollection& operator << (const char c); // Add a new character
  StringCollection& operator << (const char * str); // Add a few characters
  
  struct Ends {};
  inline friend void ends(Ends) {}	// we merely need to overload the
  					// standard ends with ours (which of
  					// a distinct type). Only the type
  					// of our ends matters, not what
  					// it does
  					
  StringCollection& operator << (void (*)(Ends)); // Terminate the string and
  					// switch the writing mode off
 
  
  ostream& dump(ostream& os);	// Dump the whole collection
  				// (for debugging purposes mostly)
  				
  		// The following class facilitates making a "shared"
  		// reference of a collection
  		// Since we lock the collection (in a shared mode)
  		// before using, it's OK to return pointers to
  		// the collection's buffer
  class CollectionRef
  {
    StringCollection& collection;
  public:
    CollectionRef(const StringCollection& _collection)
    	: collection(const_cast<StringCollection&>(_collection))
    	  { collection.get_shared(); }
    ~CollectionRef(void)	  { collection.release_shared(); }
    const char * q_end_buffer(void) const
    	{ return collection.buffer + collection.filled_size; }
    Mark q_ptr_mark(const char * a_buffer_ptr) const
    	{ const size_t offset = a_buffer_ptr - collection.buffer;
          assert( offset < collection.current_bsize ); return offset; }
    const char * q_mark_ptr(Mark mark) const
    	{ assert( mark.pos < collection.filled_size );
    	  return collection.buffer + mark.pos; }
  };
};


  		// A light-weight istrstream to access a string in
  		// a collection. STL calls the object like the
  		// following a forward iterator.
  		// This class provides a "view" of the collection,
  		// of a particular string in the collection, that is.
  		// Therefore, this object asserts a shared access to a
  		// collection
 		// This stream reaches EOF either when the actual end-of
  		// collection has seen, or if the string's terminator
  		// is reached
class iststream : protected StringCollection::CollectionRef
{
			// Note since we've got a shared lock of the collection,
			// we may use pointers now. The collection's
			// buffer can only be reallocated if the buffer
			// is held exclusive. The shared lock we hold shall
			// preclude this.
  const char * curr_ptr;
  const char * end_ptr;
  int state;		// like ios::state
    
  iststream(const iststream&);		// Deliberately unimplemented:
  void operator = (const iststream&);	// no cloning allowed!

  void setstate(ios::io_state bit)	{ state |= bit; }
  				// badbit and failbit, once set, stick!
  void clear(void)			{ state &= ~ios::eofbit; }
  
  void init(StringCollection::Mark mark);

public:
  iststream(const StringCollection& collection, const StringCollection::Mark mark);
  iststream(const Dictionary& dictionary, const char key[]);
  iststream(const Dictionary& dictionary, const StringCollection::Mark mark);
	
	// standard istream interface
  bool good(void) const		{ return state == 0; }
  bool eof(void) const		{ return state & ios::eofbit; }
  bool fail(void) const 	{ return state & (ios::badbit|ios::failbit); }
  bool bad(void) const 		{ return state & ios::badbit; }
  bool operator ! (void) const	{ return !good(); }
  
  int get(void);	// Note, reading '\0' raises (and returns) EOF!
  			// When the stream is in the EOF condition,
  			// get() and peek() crash!
  			
  			// Again, peeking() at EOF turns the stream in the
  			// EOF state. This is consistent with peek's
  			// of the regular C++ streams
  int peek(void);
  iststream& putback(const char c);
    //iststream& get(char& c);

			// Load a string dest_str (of size at most n) from this 
			// string stream
			// until EOF or one of the terminators encountered
			// The terminator is left on the stream
  iststream& getline_upto(char * dest_str, const int n,
    			  const char terminators []);
    			
    		// The same as above, but reads and discards the delimiter
  iststream& getline(char* ptr, int len, char delim = '\n');
    
    		// Checks to see if this string stream starts with 'prefix'
    		// (case is ignored)
    		// If so, position a stream at the first character
    		// after the prefix; If the prefix does not occur,
    		// set the eofbit
  iststream& begins_with_ci(const char * prefix);
  
  		// Read an integral value of the stream
  iststream& operator >> (int& val);
  
  		// Tell the current position in the stream
  		// Return the invalid mark if the stream is not in a good
  		// state
  StringCollection::Mark tellg(void) const;
};

			// The printOn operator...
ostream& operator << (ostream& os, iststream& sa);

			// Case-insensitive comparison of 'str' with a
			// string in the collection
bool strcasecmp(const char * str, iststream& sa);
bool strcasecmp(iststream& s1, iststream& s2);



		// A Dictionary of key-value pairs, with a predefined
		// search path. The items added last would be looked at
		// first during lookup.
		// Both keys and value are supposed to be strings

class DictValueHandler;
class DictKeyValueHandler;

class Dictionary
{
protected:
  mutable StringCollection buffer;	// Buffer where string bodies lie
private:
  class Assoc				// A single key-value association
  {
    friend class Dictionary;
    const StringCollection::Mark key;	// Position in buffer for header's key
    const StringCollection::Mark value;	// Position in buffer for header's value
    
    				// These two let the Dictionary construct an
    				// array of Assoc's. The array is constructed
    				// without initialization; placement
    				// operator 'new' is later called to construct
    				// an object in the already preallocated place
//#ifndef __GNUC__
    Assoc(void) : key(StringCollection::Mark::DONT_CONSTRUCT()),
                  value(StringCollection::Mark::DONT_CONSTRUCT()) {}
//#else
//    Assoc(void) {}		// Otherwise gcc crashes...
//#endif

    void * operator new(size_t, void * ptr) { return ptr; }
  public:
    Assoc(StringCollection::Mark _key, StringCollection::Mark _value)
    	: key(_key), value(_value) {}
    ostream& dump(ostream& os, StringCollection& buffer) const;
  };
  Assoc * associations;		// Ptr to the array of associations
  int no_assoc_slots;		// How many slots are in the array
  int no_assoc;			// How many of them are really valid
  
  				// We preallocate the buffer for a few
  				// associations
  Assoc assoc_buffer[7];
  
  Dictionary(const Dictionary&);	// private and unimplemented to
  void operator = (const Dictionary&);	// prohibit copy-constructing
  
protected:
				// Add a new association to the dictionary
				// (the bodies of the key and the value are
				// supposed to be in the buffer already).
  void add_assoc(const StringCollection::Mark pos_key,
  		 const StringCollection::Mark pos_value);

public:
  Dictionary(void)
  	: associations(assoc_buffer),
	  no_assoc_slots(sizeof(assoc_buffer)/sizeof(assoc_buffer[0])),
  	  no_assoc(0) {}
  ~Dictionary(void)
  	{ if( associations != assoc_buffer ) free(associations); }
 
  				// a printOn operator...
  friend ostream& operator << (ostream& os, const Dictionary& dict);
  
  				// Looking up a particular value
  				// associated with a given key
  				
  				// Lookup a value with a given key and
  				// return the mark of the value. The mark
  				// maybe invalid if the key wasn't found...
  StringCollection::Mark lookup(const char key[]) const;

  				// The same, but build the isstream out
  				// of the value. The stream would be
  				// invalid if they key was not found
  friend iststream::iststream(const Dictionary& dictionary, const char key[]);
  friend iststream::iststream(const Dictionary& dictionary,
  			      const StringCollection::Mark);
 

  				// Look up a 'key' and execute the
  				// value_taker, passing the key's value
  				// to it
  				// Return value_taker's result, or false
  				// if the key wasn't found
  bool lookup(const char key[], const DictValueHandler& value_taker) const;
  
  				// Look up a 'key' and return its value,
  				// which is converted to 'int'
  				// Return the fallback value if the conversion
  				// or the lookup failed.
  int lookup(const char key[], const int fallback) const;
  
  				// Iterate through the dictionary, see
  				// comments in the implementation file
  bool any(const DictKeyValueHandler& key_value_handler) const;
};

				// Functors used by the Dictionary functions
				// above to process a value (or a key)
struct DictValueHandler
{
  virtual bool operator() (iststream& value) const = 0;
};

struct DictKeyValueHandler
{
  virtual bool handle_key(iststream& key) const = 0;
  virtual bool handle_value(iststream& value) const = 0;
};



#endif
