/*
 ************************************************************************
 *
 *		Validate basic File System API
 *	and claims that a particular VFS (HTTPFS for one) conforms to it
 *
 * $Id: vhttpfs.c,v 1.2 1999/05/07 00:50:52 oleg Exp oleg $
 *
 ************************************************************************
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <assert.h>

static const char pattern [] = "1234\n467\r\n\007A!\177";
static const char pattern1 [] = "Yet another pattern";

                                // Make sure that a system call went well
                                // If a system call (say, fcntl()) fails
                                // it returns -1 and sets the errno
#define do_well(FX) (void)( (FX) < 0 ? (perror("System error"), \
        fprintf(stderr,"Failed system call " #FX " at line %d of `%s'.\n", \
               __LINE__, __FILE__), exit(4)) : 1)

static void do_test(const char filename[])
{
  printf("\nTesting File API using file name >%s<\n",filename);
#if 1
  { 
    int fd;
    printf("\n\tWriting a pattern to the file...");
    do_well( fd = open(filename,O_WRONLY|O_CREAT|O_TRUNC,0777) );
    assert( write(fd,pattern,sizeof(pattern)) == sizeof(pattern) );
    do_well( close(fd) );
  }
  { 
    struct stat buf;
    memset(&buf,0,sizeof(buf));
    printf("\n\tMaking sure that file we created exists and has size %d...",
	   sizeof(pattern));
    do_well( lstat(filename,&buf) );
    assert( buf.st_size == sizeof(pattern) );
    assert( S_ISREG(buf.st_mode) );
    memset(&buf,0,sizeof(buf));
    printf("\n\tthe same but with stat() rather than lstat()...");
    do_well( stat(filename,&buf) );
    assert( buf.st_size == sizeof(pattern) );
    assert( S_ISREG(buf.st_mode) );
  }
  { 
    int fd;
    char buf[sizeof(pattern)];
    printf("\n\tReading the file back...");
    do_well( fd = open(filename,O_RDONLY) );
    assert( read(fd,buf,sizeof(buf)) == sizeof(buf) );
    assert( memcmp(buf,pattern,sizeof(pattern)) == 0 );
    do_well( close(fd) );
  }
  { 
    int fd;
    printf("\n\tChecking O_CREAT|O_EXCL...");
    fd = open(filename,O_WRONLY|O_CREAT|O_EXCL,0777);
    assert( fd < 0 );
    perror("open with O_CREAT|O_EXCL predictably failed with an error");
  }
#endif
  { 
    FILE * fp;
    printf("\n\tAppending to the FILE...");
    assert( (fp = fopen(filename,"ab")) != (FILE*)0 );
    fprintf(fp, "%s", pattern1);
    do_well( fclose(fp) );
  }
  { 
    FILE * fp; register int i;
    printf("\n\tReading the appended FILE...");
    assert( (fp = fopen(filename,"rb")) != (FILE*)0 );
    for(i=0; i<sizeof(pattern); i++)
      assert( getc(fp) == pattern[i] );
    for(i=0; i<strlen(pattern1); i++)
      assert( getc(fp) == pattern1[i] );
    assert( getc(fp) == EOF );
    do_well( fclose(fp) );
  }
  { 
    FILE * fp;
    printf("\n\tTruncating the FILE to pattern1");
    assert( (fp = fopen(filename,"wb")) != (FILE*)0 );
    fprintf(fp, "%s", pattern1);
    do_well( fclose(fp) );
  }
  { 
    FILE * fp; register int i;
    printf("\n\tUpdating the FILE, over-writing the second byte");
    assert( (fp = fopen(filename,"r+b")) != (FILE*)0 );
    for(i=0; i<strlen(pattern1); i++)
      assert( getc(fp) == pattern1[i] );
    assert( getc(fp) == EOF );
    do_well( fseek(fp,1,SEEK_SET) );
    fputc('\275',fp);
    do_well( fclose(fp) );
  }
  { 
    FILE * fp; register int i;
    printf("\n\tVerifying the updated FILE");
    assert( (fp = fopen(filename,"rb")) != (FILE*)0 );
    assert( getc(fp) == pattern1[0] );
    assert( (char)getc(fp) == '\275' );
    for(i=2; i<strlen(pattern1); i++)
      assert( getc(fp) == pattern1[i] );
    assert( getc(fp) == EOF );
    do_well( fclose(fp) );
  }
  printf("\nDone\n");
}

int main(void)
{
  do_test("/tmp/x1");
  do_test("http://sun-host.com/cgi-bin/admin/MCHFS-server.pl/DeepestRoot/tmp/x2");
  do_test("http://winnt-host.com/cgi-bin/admin/MCHFS-server.pl/wwwroot/x2");
  printf("\nAll Done\n");
  return 0;
}
