// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *   Verify String collections, lightweight streams and dictionaries
 *
 * $Id: vutil.cc,v 2.2 1998/07/30 16:06:31 oleg Exp oleg $
 *
 ************************************************************************
 */

#include <iostream.h>
#include "util.h"
#include <stdlib.h>
#if defined(WIN32)
#include <string.h>
#else
#include <strings.h>
#endif

static void test_string_collection(void)
{
  cerr << "\nTesting a StringCollection" << endl;
  const char str1 [] = "string1!";
  const char str2 [] = "another s!tr";
  StringCollection collection;
  StringCollection::Mark mark1 = collection.new_string();
  cerr << "about to add >" << str1 << "< at " << mark1 << endl;
  collection << str1 << ends;
  StringCollection::Mark mark2 = collection.new_string();
  cerr << "about to add >" << str2 << "< at " << mark2 << endl;
  collection << str2 << ends;
  collection.dump(cerr);
  char buffer[100];
  assert( iststream(collection,mark1).getline_upto(buffer,sizeof(buffer),"").eof() );
  cerr << "reading at mark1: >" << buffer << "<" << endl;
  assert( strcmp(buffer,str1) == 0 );
  iststream sa(collection,mark2);
  sa.getline_upto(buffer,sizeof(buffer),"!");
  cerr << "reading at mark2: >" << buffer << "<" << endl;
  sa.getline(buffer,sizeof(buffer),'!');
  cerr << "continue reading at mark2: >" << buffer << "<" << endl;
  assert( buffer[0] == '\0' );		// The buffer must be empty
  sa.getline(buffer,sizeof(buffer),'!');
  cerr << "continue reading at mark2: >" << buffer << "<" << endl;
  assert( sa.eof() );
  cerr << "re-reading at mark1: >" << iststream(collection,mark1) << "<" << endl;
  cerr << "\nDone\n" << endl;
}

static void test_string_comparisons(void)
{
  cerr << "\nValidating case-independent string comparisons" << endl;
  
  struct TestCase { const char * full_string;
  		    const char * prefix;
  	   	    bool expected_result; }
  test_cases [] =
  	{ {"abcdef ","abc", true},	{" abcdef "," abc", true},
  	  {"abc","abc", true}, 		{"aBc","AbC", true},
  	  {"aBc\n","AbC", true},	{"aBe\n","AbC",false},
  	  {"acc","abc", false},		{"ab","abc", false},
  	  {"a123456789.,a","A123456789.,", true},
  	  {"a1234bB?\\def","A1234Bb?\\", true},
  	  {"\rZ!@~#$Ll*()def","\rz!@~#$lL*()", true},
  	  {"a1234bB.\\def","A1234Bb?\\", false}
  	};

  cerr << "\tloading the test strings into a dictionary,"
       << "\n\twith keys identical to their values" << endl;

  class TestDict : public Dictionary
  {
  public:
    TestDict(TestCase cases [], const int no_cases)
    {
      for(register int i=0; i<no_cases; i++)
      {
        StringCollection::Mark mark_key = buffer.new_string();
        buffer << cases[i].full_string << ends;
        StringCollection::Mark mark_val = buffer.new_string();
        buffer << cases[i].full_string << ends;
        add_assoc(mark_key,mark_val);
      }
    }
    bool test_string(const char str [], const char prefix [])
     {
       cerr << "looking for '" << prefix << "' in '" << str << "': ";
       iststream coll_str(*this,str);
       if( !(coll_str.begins_with_ci(prefix)) )
         return (cerr << " #f" << endl), false;
       const int c = coll_str.get();
       cerr << " #t; next char ";
       if( c == EOF )
         cerr << "EOF";
       else
         cerr << '\'' << (char)c << '\'';
       cerr << endl;
       return true;
     }
  };
  
  TestDict test_dict(test_cases,sizeof(test_cases)/sizeof(test_cases[0]));
  for(register size_t i=0; i<sizeof(test_cases)/sizeof(test_cases[0]); i++)
  {
    const TestCase& test_case = test_cases[i];
    assert( test_dict.test_string(test_case.full_string,test_case.prefix) ==
    	    test_case.expected_result );
  }
  cerr << "\nDone\n" << endl;
}

static void test_int_value_taking(void)
{
  cerr << "\nValidating reading int values" << endl;
  
  struct TestCase { const char * full_string;
  		    int default_value;
  	   	    int expected_result; }
  test_cases [] =
  	{ {"", -1, -1},		{" ", -1, -1},
  	  {"+", -1, -1},	{" - ", 1, 1},
  	  {"0", -1, 0},		{"-0", -1, 0},
  	  {"+0", -1, 0},	{"++0", -1, -1},
  	  {"1234", -1, 1234},	{"+1234", -1, 1234},
  	  {"-1234000", -1, -1234000},	{"+1-234", -1, 1},
  	  {" - 1234 000", -1, -1234},	{"+ -234", -1, -1},
  	};

  cerr << "\tloading the test strings into a dictionary,"
       << "\n\twith keys identical to their values" << endl;

  class TestDict : public Dictionary
  {
  public:
    TestDict(TestCase cases [], const int no_cases)
    {
      for(register int i=0; i<no_cases; i++)
      {
        StringCollection::Mark mark_key = buffer.new_string();
        buffer << cases[i].full_string << ends;
        StringCollection::Mark mark_val = buffer.new_string();
        buffer << cases[i].full_string << ends;
        add_assoc(mark_key,mark_val);
      }
    }
    int test_conv(const char key [], const int default_value)
     {
       cerr << "converting '" << key << "' into an int, default "
            << default_value << ": ";
       const int result = lookup(key,default_value);
       cerr << result << endl;
       return result;
     }
  };
  
  TestDict test_dict(test_cases,sizeof(test_cases)/sizeof(test_cases[0]));
  for(register size_t i=0; i<sizeof(test_cases)/sizeof(test_cases[0]); i++)
  {
    const TestCase& test_case = test_cases[i];
    assert( test_dict.test_conv(test_case.full_string,test_case.default_value) ==
    	    test_case.expected_result );
  }
  cerr << "\nDone\n" << endl;
}



main(void)
{
  test_string_collection();
  test_string_comparisons();
  test_int_value_taking();
  
  cerr << "\nAll tests passed\n" << endl;
}
